/*
 * $QNXLicenseC: 
 * Copyright 2007, QNX Software Systems.  
 *  
 * Licensed under the Apache License, Version 2.0 (the "License"). You  
 * may not reproduce, modify or distribute this software except in  
 * compliance with the License. You may obtain a copy of the License  
 * at: http://www.apache.org/licenses/LICENSE-2.0  
 *  
 * Unless required by applicable law or agreed to in writing, software  
 * distributed under the License is distributed on an "AS IS" basis,  
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied. 
 * 
 * This file may contain contributions from others, either as  
 * contributors under the License or as licensors under other terms.   
 * Please review this entire file for other proprietary rights or license  
 * notices, as well as the QNX Development Suite License Guide at  
 * http://licensing.qnx.com/license-guide/ for other information. 
 * $ 
 */




/*

PCI device configuration manager

*/

struct _dev_list;
#define IOFUNC_OCB_T	struct _dev_list

#include	<unistd.h>
#include	<stdio.h>
#include	<stdarg.h>
#include	<stdlib.h>
#include	<stddef.h>
#include	<string.h>
#include	<errno.h>
#include	<signal.h>
#include	<dl.h>
#include	<pthread.h>
#include	<gulliver.h>
#include	<sys/neutrino.h>
#include	<sys/iofunc.h>
#include	<sys/iomgr.h>
#include	<sys/iomsg.h>
#include	<sys/dispatch.h>
#include	<sys/rsrcdbmgr.h>
#include	<pcidrvr.h>
#include	<server.h>
#include	<sys/procmgr.h>
#include	<sys/slogcodes.h>

#define		MAX_BUSES	10
#define		NUM_DEVICES	5
#define		SERV_MAX_IRQS	16
#define		DEFAULT_SIZE	256
#define		MEG1			(1024 * 1024)
#define		KB4				(4 * 1024)

#define	CARDBUS_PRI_BUS		offsetof (struct _pci_cardbus_config_regs, Pci_Bus_Num)
#define	CARDBUS_BUS			offsetof (struct _pci_cardbus_config_regs, Cardbus_Bus_Num)
#define	CARDBUS_SUB_BUS		offsetof (struct _pci_cardbus_config_regs, Sub_Bus_Num)

union _pci_data {
	struct _pci_present		present;
	struct _pci_config		config;
	struct _pci_config_hdl	config_hdl;
	struct _pci_device		device;
	struct _pci_class		class;
	struct _pci_map			map;
	struct _pci_version		version;
	struct _pci_attach		lock;
	struct _pci_detach		unlock;
	struct _pci_map_irq		map_irq;
	struct _pci_route_opt	route_opt;
};

static	const	pci_bus_t	_x86_pci_bus = {
	0,			/* minimum grant */
	255,		/* maximum latency */
	0,			/* devsel time = fast */
	1,			/* we support fast back-to-back */
	1,			/* we support prefetch */
	0,			/* we don't support 66 Mhz */
	0,			/* we don't support 64 bits */
	1,			/* initially one device on bus (ie. us) */
	132000000	/* bandwidth: in 0.25us cycles/sec */
};

typedef	struct _dev_list {
	iofunc_ocb_t	ocb;
	int32_t			num_devices, lockflg;
	Device			*dev_list [NUM_DEVICES];
	} DEV_LIST;

typedef struct _pdrvr {
	pdrvr_entry_t	*entry;
	void			*hdl;
	void 			*next;
	} pdrvr_t;

typedef	struct {
	uint64_t		io_reqd;
	uint64_t		mem_reqd;
	uint64_t		pref_reqd;
	} bus_require_t;

pdrvr_t		*pci_drivers;
static		pthread_mutex_t		serv_mutex = PTHREAD_MUTEX_INITIALIZER;

static		int32_t		numdevices, listcount;
static		int32_t		scan_buses;
static		Device		*DeviceHead, *DeviceTail;
static		BUSMAP		*bus_map;
static		uint32_t	version, lastbus, hardware;
static		uint32_t	enum_xtrans_bridge = 0;
static		int32_t		dont_touch = 0, hw_override = 0;
static		int32_t		program_display_device = 0;
static		uint64_t	mem_cpu_low, mem_cpu_high, mem_pci_low, mem_pci_high;
static		uint64_t	io_cpu_low, io_cpu_high, io_pci_low, io_pci_high;
static		uint8_t		host_bus = 0xff;
static		uint32_t	host_devfunc;
static		uint32_t	host_bridge_class = 0xffffffff;

int32_t		verbose = 0;

static		resmgr_io_funcs_t io_funcs;

static		void		*ll_handle;

static		int8_t		*progname = "pci_server:";

static	int32_t		pci_free_resource (struct BaseRegister *BaseReg, uint8_t bus);
static	uint32_t	pci_free_irq (uint32_t irq);
static	int32_t		scan_windows (uint8_t BusNumber, uint32_t DevFuncNumber, Device *device);

static	int32_t		map_irqs = 1;

static	int		cur_scan_bus;

static	bus_require_t	brqt [20];

/* pci calls */

/***********************************************************************/
/*                                                                     */
/***********************************************************************/

ssize_t		my_slogf (int opcode, int severity, char *fmt, ...)

{
ssize_t		ret;
va_list		arglist;

	if (verbose > 4) {
		va_start (arglist, fmt);
		vfprintf (stderr, fmt, arglist);
		fprintf (stderr, "\n");
		va_end (arglist);
		}
	va_start (arglist, fmt);
	ret = vslogf (opcode, severity, fmt, arglist);
	va_end (arglist);
	return (ret);
}

/**************************************************************************/
/* This routine currently plugs in fixed values into the version and      */
/* hardware fields. The lastbus field will contain the number of the last */
/* PCI bus found during the bus scan.                                     */
/**************************************************************************/

static	int32_t		pci_bios_present (uint32_t *last_bus, uint32_t *vers, uint32_t *hardw)

{
	if (!lastbus) {		//No PCI found
		return (PCI_UNSUPPORTED_FUNCT);
		}
	*last_bus = lastbus;
	*vers = version;
	*hardw = hardware;
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* All calls to the low level PCI routines will read/write 32-bit         */
/* quantities. The following 6 routines will extract the relevant         */
/* information from these 32-bit fields.                                  */
/**************************************************************************/

static	int		PCI_read_config8 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, offset, cnt, status = 0;
uint8_t		*p;
uint8_t		temp [256];

	if (Cnt > sizeof (temp))
		return (PCI_BUFFER_TOO_SMALL);
	pthread_mutex_lock (&serv_mutex);
	offset = Offset & ~3;
	cnt = (Cnt+3) >> 2;
	if (cnt > 64)
		cnt = 64;
	for (i = 0, p = temp; i < cnt; i++, p += sizeof (ulong_t), offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->read_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, p)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "read_config8 status %x", status);
			return (status);
			}
		}
	memcpy (ConfigBuffer, &temp [Offset % 4], Cnt);
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* See PCI_read_config8                                                   */
/**************************************************************************/

static	int		PCI_read_config16 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, offset, cnt, status = 0;
uint16_t	*p;
uint8_t		temp [256];
uint32_t	dword;

	if ((Cnt * 2) > sizeof (temp))
		return (PCI_BUFFER_TOO_SMALL);
	if (Offset & 1)			//Can't have odd offsets
		return (PCI_BAD_REGISTER_NUMBER);
	pthread_mutex_lock (&serv_mutex);
	offset = Offset & ~3;
	cnt = (Cnt+1) >> 1;
	if (cnt > 64)
		cnt = 64;
	for (i = 0, p = (uint16_t *) temp; i < cnt; i++, offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->read_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, &dword)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "read_config16 status %x", status);
			return (status);
			}
		dword = ENDIAN_LE32 (dword);
		*p++ = dword & 0xffff;
		*p++ = (dword & 0xffff0000) >> 16;
		}
	memcpy (ConfigBuffer, &temp [Offset % 4], Cnt * 2);
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* See PCI_read_config8                                                   */
/**************************************************************************/

static	int		PCI_read_config32 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, status = 0;
uint32_t	*p;

	if (Offset & 3)			//Can't have odd offsets
		return (PCI_BAD_REGISTER_NUMBER);
	pthread_mutex_lock (&serv_mutex);
	for (i = 0, p = ConfigBuffer; i < Cnt; i++, p++, Offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->read_cnfg) (ll_handle, BusNumber, DevFuncNumber, Offset, 4, (uint8_t *) p)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "read_config32 status %x", status);
			return (status);
			}
		*p = ENDIAN_LE32 (*p);
		}
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* The PCI write routines are similar to the read functions, except that  */
/* where a value less than 32-bits is written, a read will first have to  */
/* be performed and the new value ored into this 32-bit value.            */
/**************************************************************************/

static	int		PCI_write_config8 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, offset, cnt, status = 0;
uint8_t		*p;
uint8_t		temp [256];

	if (Cnt > sizeof (temp))
		return (PCI_BUFFER_TOO_SMALL);
	pthread_mutex_lock (&serv_mutex);
	offset = Offset & ~3;
	cnt = (Cnt+3) >> 2;
	if (cnt > 64)
		cnt = 64;
	for (i = 0, p = temp; i < cnt; i++, p += sizeof (ulong_t), offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->read_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, p)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "write_config8 status %x", status);
			return (status);
			}
		}
	memcpy (&temp [Offset % 4], ConfigBuffer, Cnt);
	offset = Offset & ~3;
	for (i = 0, p = temp; i < cnt; i++, p += sizeof (ulong_t), offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->write_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, (*(uint32_t *) p))) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			return (status);
			}
		}
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* See PCI_write_config8                                                  */
/**************************************************************************/

static	int		PCI_write_config16 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, offset, cnt, status = 0;
uint16_t	*p, *q;
uint8_t		temp [256];
uint32_t	dword;

	if ((Cnt * 2) > sizeof (temp))
		return (PCI_BUFFER_TOO_SMALL);
	if (Offset & 1)			//Can't have odd offsets
		return (PCI_BAD_REGISTER_NUMBER);
	pthread_mutex_lock (&serv_mutex);
	offset = Offset & ~3;
	cnt = (Cnt+1) >> 1;
	if (cnt > 64)
		cnt = 64;
	for (i = 0, p = (uint16_t *) temp; i < cnt; i++, offset += sizeof (ulong_t)) {
		if ((status = (*pci_drivers->entry->read_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, &dword)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "write_config16 status %x", status);
			return (status);
			}
		dword = ENDIAN_LE32 (dword);
		*p++ = dword & 0xffff;
		*p++ = (dword & 0xffff0000) >> 16;
		}
	for (i = 0, p = (uint16_t *) &temp [Offset % 4], q = (uint16_t *) ConfigBuffer; i < Cnt; i++)
		*p++ = *q++;
	offset = Offset & ~3;
	for (i = 0, p = (uint16_t *) temp; i < cnt; i++, offset += sizeof (ulong_t)) {
		dword = *p++;
		dword |= *p++ << 16;
		dword = ENDIAN_LE32 (dword);
		if ((status = (*pci_drivers->entry->write_cnfg) (ll_handle, BusNumber, DevFuncNumber, offset, 4, dword)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			return (status);
			}
		}
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* See PCI_write_config8                                                  */
/**************************************************************************/

static	int		PCI_write_config32 (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t Cnt, void *ConfigBuffer)

{
int			i, status = 0;
const		uint32_t	*p;
uint32_t	dword;

	if (Offset & 3)			//Can't have odd offsets
		return (PCI_BAD_REGISTER_NUMBER);
	pthread_mutex_lock (&serv_mutex);
	for (i = 0, p = ConfigBuffer; i < Cnt; i++, p++, Offset += sizeof (ulong_t)) {
		dword = ENDIAN_LE32 (*p);
		if ((status = (*pci_drivers->entry->write_cnfg) (ll_handle, BusNumber, DevFuncNumber, Offset, 4, dword)) != PCI_SUCCESS) {
			pthread_mutex_unlock (&serv_mutex);
			if (verbose > 3)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "write_config32 status %x", status);
			return (status);
			}
		}
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* Turn on access to the PCI device                                       */
/**************************************************************************/

static void	on_pci (uint8_t BusNumber, uint32_t DevFuncNumber, uint16_t flags)

{
uint16_t	cr;

	if (dont_touch)
		return;
	PCI_read_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Command), 1, (char *)&cr);
	cr |= flags;
	PCI_write_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Command), 1, (char *)&cr);
}

/**************************************************************************/
/* Turn off access to the PCI device, except for configuration            */
/**************************************************************************/

static uint16_t	off_pci (uint8_t BusNumber, uint32_t DevFuncNumber)

{
uint16_t	cr, ret;

	PCI_read_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Command), 1, (char *)&cr);
	ret = cr;
	if (dont_touch)
		return (ret);
	cr &= ~3;
	PCI_write_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Command), 1, (char *)&cr);
	return (ret);
}

/**************************************************************************/
/* Check device capabilities                                              */
/**************************************************************************/

static int	check_capabilities (uint8_t BusNumber, uint32_t DevFuncNumber, Device *device)

{
uint8_t		cap_ptr, cap;
uint16_t	pwr;
int			capab;
	
	if ((device->HeaderType & 0x7f) == 2)	/* CardBus device */
		capab = offsetof (struct _pci_cardbus_config_regs, Capabilities_Pointer);
	else
		capab = offsetof (struct _pci_config_regs, Capabilities_Pointer);
	if (PCI_read_config8 (BusNumber, DevFuncNumber, capab, 1, &cap_ptr) != PCI_SUCCESS)
		return (-1);

	if (cap_ptr) {

		while (1) {
			if (PCI_read_config8 (BusNumber, DevFuncNumber, cap_ptr, 1, &cap) != PCI_SUCCESS)
				return (-1);
			if (cap == PCI_CAP_PCIX)
				device->PciXDev = 1;
			if (cap == PCI_CAP_POWER_MANAGEMENT)
				break;
			if (PCI_read_config8 (BusNumber, DevFuncNumber, cap_ptr + 1, 1,
				&cap_ptr) != PCI_SUCCESS)
				return (-1);
			if (!cap_ptr)
				return (EOK);
			}
				
		if (PCI_read_config16 (BusNumber, DevFuncNumber, cap_ptr + 4, 1, &pwr) != PCI_SUCCESS)
			return (-1);

		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Bus %x devfunc %x pwr %x", BusNumber,
				DevFuncNumber, pwr);
		device->PmState = pwr;
		device->PmCap = cap_ptr;
		}

	return (EOK);
}

static	BUSMAP	*find_bus (uint8_t bus);

/**************************************************************************/
/* Allocate a new pci bus structure every time a new bus is discovered.   */
/**************************************************************************/

static	BUSMAP	*alloc_bus (uint8_t pribus, uint8_t secbus)

{
pci_bus_t	*bus;
BUSMAP		*new_map, *map = bus_map;

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "alloc_bus %d %d", pribus, secbus);

	if ((bus = calloc (1, sizeof (pci_bus_t))) == NULL) {
		return (NULL);
		}
	if ((new_map = calloc (1, sizeof (BUSMAP))) == NULL) {
		free (bus);
		return (NULL);
		}

	if (map == NULL) {
		bus_map = new_map;
		}
	else {
		if ((map = find_bus (pribus)) == NULL) {
			free (bus);
			free (new_map);
			return (NULL);
			}
		new_map->prev = map;
		if (map->next == NULL) {
			map->next = new_map;
			}
		else {
			map = map->next;
			if (map->primary_bus == pribus) {
				if (map->peer == NULL) {
					map->peer = new_map;
					}
				else {
					while (map->peer) {
						map = map->peer;
						}
					map->peer = new_map;
					}
				}
			}
		}
	new_map->pci_bus = bus;
	*new_map->pci_bus = _x86_pci_bus;		/* initialise the structure */
	new_map->primary_bus = pribus;
	new_map->secondary_bus = secbus;
	new_map->subordinate_bus = lastbus++;
	sprintf (new_map->name, "bus%02d", secbus);
	return (new_map);
}

/**************************************************************************/
/* Find the bus structure associated with a particular bus number and     */
/* return a pointer to this structure.                                    */
/**************************************************************************/

static	BUSMAP	*find_bus (uint8_t bus)

{
BUSMAP	*map = bus_map;

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "find_bus %d", bus);
	while (1) {
		if (map == NULL) {
			break;
			}
		if (map->secondary_bus == bus) {
			return (map);
			}
		if (map->secondary_bus) {
			if (bus > map->subordinate_bus) {
				if (map->peer) {
					map = map->peer;
					continue;
					}
				}
			}
		if (map->next) {
			map = map->next;
			continue;
			}
		if (map->peer == NULL) {
			break;
			}
		map = map->peer;
		}
	return (NULL);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	void	find_ranges (void)

{
rsrc_request_t	rreq;
uint64_t		map_addr;

	mem_cpu_low = mem_cpu_high = mem_pci_low = mem_pci_high = 0;
	io_cpu_low = io_cpu_high = io_pci_low = io_pci_high = 0;

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = 1;
	rreq.start = rreq.end = 0;
	rreq.flags = RSRCDBMGR_PCI_MEMORY;
	rsrcdbmgr_attach (&rreq, 1);
	(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_MEM_SPACE | PCI_MAP_ADDR_PCI);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Low memory CPU %llx - PCI %llx", rreq.start, map_addr);
	mem_cpu_low = rreq.start;
	mem_pci_low = map_addr;
	rsrcdbmgr_detach (&rreq, 1);

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = 1;
	rreq.start = rreq.end = 0;
	rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_TOPDOWN;
	rsrcdbmgr_attach (&rreq, 1);
	(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_MEM_SPACE | PCI_MAP_ADDR_PCI);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "High memory CPU %llx - PCI %llx", rreq.start - 1, map_addr);
	mem_cpu_high = rreq.start;
	mem_pci_high = map_addr;
	rsrcdbmgr_detach (&rreq, 1);

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = 1;
	rreq.start = rreq.end = 0;
	rreq.flags = RSRCDBMGR_IO_PORT;
	rsrcdbmgr_attach (&rreq, 1);
	(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_IO_SPACE | PCI_MAP_ADDR_PCI);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Low port CPU %llx - PCI %llx", rreq.start, map_addr);
	io_cpu_low = rreq.start;
	io_pci_low = map_addr;
	rsrcdbmgr_detach (&rreq, 1);

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = 1;
	rreq.start = rreq.end = 0;
	rreq.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_TOPDOWN;
	rsrcdbmgr_attach (&rreq, 1);
#ifdef	__X86__
	if (rreq.start > 0xffb0) {
		rreq.start = 0xffb0;
		}
#endif
	(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_IO_SPACE | PCI_MAP_ADDR_PCI);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "High port CPU %llx - PCI %llx", rreq.start - 1, map_addr);
	io_cpu_high = rreq.start;
	io_pci_high = map_addr;
	rsrcdbmgr_detach (&rreq, 1);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

int		check_range (uint64_t start, uint64_t length, int type, uint8_t bus)

{
uint64_t	low, high;
BUSMAP		*map = NULL;

	low = high = 0;
	if (bus) {
		if ((map = find_bus (bus)) == NULL) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "CR bus not found %d", bus);
			return (EINVAL);
			}
		}

	if (type & PCI_IO_SPACE) {
		if (bus && (map->io_start || map->io_end)) {
			low = map->io_start;
			high = map->io_end;
			}
		else {
			low = io_pci_low;
			high = io_pci_high;
			}
		}
	else {
		if (bus) {
			if (type & PCI_MEM_PREFETCH) {
				low = map->prefetch_mem_start;
				high = map->prefetch_mem_end;
				}
			else {
				low = map->mem_start;
				high = map->mem_end;
				}
			}
//		else {
		if (!bus || (!low && !high)) {
			low = mem_pci_low;
			high = mem_pci_high;
			}
		}
		
	if (start >= low && ((start + length - 1) <= high))
		return (0);

#ifdef	__X86__		//Only temporary - must be fixed!!!
	if (type & PCI_ROM_SPACE) {
		if (start >= 0xa0000 && ((start + length - 1) <= 0xfffff))
			return (0);
		}
#endif

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Range failed %llx - %llx", start, length);

	return (1);
}

/**************************************************************************/
/* Allocate a memory or I/O resource from the resource manager and place  */
/* the result in the BaseReg structure.                                   */
/**************************************************************************/

static	int		pci_alloc_resource (struct BaseRegister *BaseReg, uint32_t align, uint8_t bus)

{
rsrc_request_t	rreq;
BUSMAP			*busmap = NULL;
char			tmp [20];

	if (BaseReg->Flags & PCI_DEVICE_ENABLED) {	//Don't alloc a resource if device enabled
		return (PCI_SUCCESS);
		}

	if (bus) {		//We only need the bus structure for buses > 0
		if ((busmap = find_bus (bus)) == NULL) {
			return (EINVAL);
			}
		}

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = BaseReg->Size;
	if (BaseReg->Flags & PCI_MEM_SPACE || BaseReg->Flags & PCI_ROM_SPACE || BaseReg->Flags & PCI_MEM_PREFETCH) {
		BaseReg->Address &= 0x0f;		//Keep flags
		if (bus) {
			if (BaseReg->Flags & PCI_MEM_PREFETCH) {
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->prefetch_mem_start, &rreq.start, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->prefetch_mem_end, &rreq.end, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
				}
			else {
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->mem_start, &rreq.start, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->mem_end, &rreq.end, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
				}
			}
		else {
			rreq.start = rreq.end = 0;
			}
		rreq.align = align ? align : rreq.length;
		if (rreq.start == rreq.end) {
			rreq.start = rreq.end = 0;		//ppc mapped addresses are non-zero!
			rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_TOPDOWN;
			}
		else {
			rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_RANGE;
			}
		if (verbose > 2)
			sprintf (tmp, "Memory ");
		}
	else {
		if (BaseReg->Flags & PCI_IO_SPACE) {
			BaseReg->Address &= 0x3;		//Keep flags
			if (bus) {
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->io_start, &rreq.start, PCI_IO_SPACE | PCI_MAP_ADDR_CPU);
				(*pci_drivers->entry->map_addr) (ll_handle, busmap->io_end, &rreq.end, PCI_IO_SPACE | PCI_MAP_ADDR_CPU);
				}
			else {
				rreq.start = rreq.end = 0;
				}
			rreq.align = align ? align : rreq.length;
			if (rreq.start == rreq.end) {
				rreq.start = io_cpu_low;
				rreq.end = io_cpu_high;
				rreq.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_TOPDOWN | RSRCDBMGR_FLAG_RANGE;
				}
			else {
				rreq.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_RANGE;
				}
			if (verbose > 2)
				sprintf (tmp, "Port ");
			}
		else {
			return (EINVAL);
			}
		}
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "%sRequest Start %llx - End %llx - Size = %llx", tmp, rreq.start, rreq.end, rreq.length);
	if (rsrcdbmgr_attach (&rreq, 1)) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Failed %s", strerror (errno));
		return (EINVAL);
		}

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Address = %llx - End %llx", rreq.start, rreq.end);
	BaseReg->Address |= rreq.start;		//Or address with flags
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Free a previously allocated memory or I/O resource                     */
/**************************************************************************/

static	int		pci_free_resource (struct BaseRegister *BaseReg, uint8_t bus)

{
rsrc_request_t	rreq;
BUSMAP			*busmap = NULL;
char			tmp [20];

	if (bus) {		//We only need the bus name for buses > 0
		if ((busmap = find_bus (bus)) == NULL) {
			return (EINVAL);
			}
		}

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = BaseReg->Size < 256 ? 256 : BaseReg->Size;
	rreq.name = busmap ? busmap->name : NULL;
	if (BaseReg->Flags & PCI_MEM_SPACE || BaseReg->Flags & PCI_ROM_SPACE || BaseReg->Flags & PCI_MEM_PREFETCH) {
		rreq.start = PCI_MEM_ADDR (BaseReg->Address);
		rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_RANGE;
		BaseReg->Address &= 0x0f;		//Keep flags
		if (verbose > 2)
			sprintf (tmp, "Free Memory ");
		}
	else {
		if (BaseReg->Flags & PCI_IO_SPACE) {
			rreq.start = PCI_IO_ADDR (BaseReg->Address);
			rreq.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_RANGE;
			BaseReg->Address &= 0x3;		//Keep flags
			if (verbose > 2)
				sprintf (tmp, "Free Port ");
			}
		else {
			return (EINVAL);
			}
		}
	rreq.end = rreq.start + rreq.length - 1;
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "%sStart %llx - End %llx - Length %llx",
					tmp, rreq.start, rreq.end, rreq.length);
	if (rsrcdbmgr_detach (&rreq, 1)) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Failed");
		return (EINVAL);
		}

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Address = %llx", rreq.start);
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Request an IRQ from the resource manager. Step through list for cnt    */
/* items until an IRQ can be allocated.                                   */
/**************************************************************************/

static	uint32_t	pci_alloc_irq (uint32_t	*list, int cnt, uint8_t bus)

{
rsrc_request_t	*rreq, *req1;
int				i;
uint32_t		irq;

	if ((rreq = calloc (cnt, sizeof (rsrc_request_t))) == NULL) {
		return (0);
		}
	for (i = 0, req1 = rreq; i < (cnt & 0xff); i++, req1++) {
		req1->length = 1;
		req1->start = *list;
		req1->end = *list++;
		req1->flags = RSRCDBMGR_IRQ | RSRCDBMGR_FLAG_RANGE | ( cnt == 1 ? RSRCDBMGR_FLAG_SHARE : 0 );
		if (cnt && i < ((cnt & 0xff) - 1))
			req1->flags |= RSRCDBMGR_FLAG_LIST;
		}

	if (rsrcdbmgr_attach (rreq, cnt) != -1) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Alloc IRQ %lld - Flags %x", rreq->start, rreq->flags);
		irq = (uint32_t) rreq->start;
		free (rreq);
		return (irq);
		}

	if (verbose > 2) {
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "IRQ Failed %lld - Flags %x", rreq->start, rreq->flags);
		}
	free (rreq);
	return (0);
}

/**************************************************************************/
/* Return an IRQ to the resource manager.                                 */
/**************************************************************************/

static	uint32_t	pci_free_irq (uint32_t irq)

{
rsrc_request_t	rreq;

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.length = 1;
	rreq.start = irq;
	rreq.end = irq;
	rreq.flags = RSRCDBMGR_IRQ | RSRCDBMGR_FLAG_RANGE;
	if (! rsrcdbmgr_detach (&rreq, 1)) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Free IRQ %d", irq);
		return (rreq.start);
		}
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "IRQ Free Failed");
	return (0);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/
#ifndef	__PPC__

static	int		pci_reserve_irq (uint32_t irq)

{
rsrc_alloc_t	ralloc;

	memset ((char *) &ralloc, 0, sizeof (rsrc_alloc_t));
	ralloc.start = ralloc.end = irq;
	ralloc.flags = RSRCDBMGR_IRQ | RSRCDBMGR_FLAG_RSVP;
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "reserve IRQ %lld", ralloc.start);
	return (rsrcdbmgr_create (&ralloc, 1));
}
#endif

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	int		configure_bridge (uint8_t bus, uint32_t devfunc)

{
uint64_t		qword;
uint32_t		dword, dword1;
uint16_t		word, word1;
BUSMAP			*map;
rsrc_alloc_t	ralloc;
#if	0
uint8_t			buf [256];
#endif

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Configure bridge - bus %d - devfunc %x", bus, devfunc);
	if ((map = find_bus (lastbus)) == NULL) {
		if ((map = alloc_bus (bus, lastbus)) == NULL)
			return (-1);
		}
	if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
		return (-1);
	dword >>= 8;
	if ((dword & 0xffff00) == (PCI_CLASS_BRIDGE)) {
		return (PCI_SUCCESS);
		}

	PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
	dword = PCI_PRIMARY_BUS_INSERT (dword, bus);
	dword = PCI_SECONDARY_BUS_INSERT (dword, lastbus - 1);
	dword = PCI_SUBORDINATE_BUS_INSERT (dword, 0xff);	//Allow config. cycles to propagate
	PCI_write_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
	PCI_read_config16 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Command), 1, &word);
	word |= PCI_COMMAND_MASTER_ENABLE;
	PCI_write_config16 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Command), 1, &word);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "cb Bus %x - Devfn %x - dword %x", bus, devfunc, dword);
	if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
		return (-1);
	dword >>= 8;
	if ((dword & 0xffff00) == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_PCI)) {
		if ((PCI_read_config16 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, IO_Base), 1, (char *)&word) != PCI_SUCCESS))
			return (-1);
		word1 = 0;
		if (word & 0x0f) {		//32-bit I/O addressing
			if ((PCI_read_config16 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, IO_Base_Upper16), 1, (char *)&word1) != PCI_SUCCESS))
				return (-1);
			}
		map->io_start = ((uint64_t)(word1 << 16) | ((uint64_t)(word & 0xf0) << 8));
		word1 = 0;
		if (word & 0x0f00) {		//32-bit I/O addressing
			if ((PCI_read_config16 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, IO_Limit_Upper16), 1, (char *)&word1) != PCI_SUCCESS))
				return (-1);
			}
		map->io_end = (((uint64_t)(word1 << 16) | (uint64_t)(word & 0xf000)) + 0xfff);
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "bus %d iostart %llx - ioend %llx", bus, map->io_start, map->io_end);
		if (map->io_end < map->io_start) {		// I/O not used
			map->io_start = map->io_end = 0;
			}
		if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Memory_Base), 1, (char *)&dword) != PCI_SUCCESS))
			return (-1);
		map->mem_start = (uint64_t) (dword & 0xffff) << 16;
		map->mem_end = (uint64_t) ((dword & 0xfff00000) + 0xfffff);
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "bus %d memstart %llx - memend %llx", bus, map->mem_start, map->mem_end);
		if (map->mem_end < map->mem_start) {	// Memory not used
			map->mem_start = map->mem_end = 0;
			}
		if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Prefetchable_Memory_Base), 1, (char *)&dword) != PCI_SUCCESS))
			return (-1);
		qword = 0;
		if (dword & 0x0f) {		//64-bit addressing
			if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Prefetchable_Base_Upper32), 1, (char *)&dword1) != PCI_SUCCESS))
				return (-1);
			qword = (uint64_t) dword1;
			}
		map->prefetch_mem_start = (uint64_t) ((qword << 32) | ((dword & 0xfff0) << 16));
		qword = 0;
		if (dword & 0xf0000) {	//64-bit addressing
			if ((PCI_read_config32 (bus, devfunc, offsetof (struct _pci_bridge_config_regs, Prefetchable_Limit_Upper32), 1, (char *)&dword1) != PCI_SUCCESS))
				return (-1);
			qword = (uint64_t) dword1;
			}
		map->prefetch_mem_end = (uint64_t) (((qword << 32) | (dword & 0xfff00000)) + 0xfffff);
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "bus %d prefstart %llx - prefend %llx", bus, map->prefetch_mem_start, map->prefetch_mem_end);
		if (map->prefetch_mem_end < map->prefetch_mem_start) {	// Memory not used
			map->prefetch_mem_start = map->prefetch_mem_end = 0;
			}
		if (map->io_start) {
			memset ((char *) &ralloc, 0, sizeof (ralloc));
			(*pci_drivers->entry->map_addr) (ll_handle, map->io_start, &ralloc.start, PCI_IO_SPACE | PCI_MAP_ADDR_CPU);
			(*pci_drivers->entry->map_addr) (ll_handle, map->io_end, &ralloc.end, PCI_IO_SPACE | PCI_MAP_ADDR_CPU);
			ralloc.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_RSVP;
			if (rsrcdbmgr_create (&ralloc, 1) == -1) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Reserve IO failed");
				}
			}
		if (map->mem_start) {
			memset ((char *) &ralloc, 0, sizeof (ralloc));
			(*pci_drivers->entry->map_addr) (ll_handle, map->mem_start, &ralloc.start, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
			(*pci_drivers->entry->map_addr) (ll_handle, map->mem_end, &ralloc.end, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
			ralloc.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_RSVP;
			if (rsrcdbmgr_create (&ralloc, 1) == -1) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Reserve MEM failed");
				}
			}
		if (map->prefetch_mem_start) {
			memset ((char *) &ralloc, 0, sizeof (ralloc));
			(*pci_drivers->entry->map_addr) (ll_handle, map->prefetch_mem_start, &ralloc.start, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
			(*pci_drivers->entry->map_addr) (ll_handle, map->prefetch_mem_end, &ralloc.end, PCI_MEM_SPACE | PCI_MAP_ADDR_CPU);
			ralloc.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_RSVP;
			if (rsrcdbmgr_create (&ralloc, 1) == -1) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Reserve RMEM failed - Start %llx - End %llx", ralloc.start, ralloc.end);
				}
			}
		}

	return (PCI_SUCCESS);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	int		Setup_CardBus_Bridge (uint8_t Bus, uint32_t DevFn, Device *Dev)

{
uint8_t		tmp, sub_bus = 0;
BUSMAP		*map;
uint16_t	status;
uint32_t	dword;


	tmp = Bus;
	if (PCI_write_config8 (Bus, DevFn, CARDBUS_PRI_BUS, 1, (int8_t *) &tmp) != PCI_SUCCESS) {
		return (-1);
		}
	if (PCI_read_config8 (Bus, DevFn, CARDBUS_BUS, 1, (int8_t *) &tmp) != PCI_SUCCESS) {
		return (-1);
		}
	if (tmp == 0 || tmp >= lastbus) {
		sub_bus = lastbus;
		if (PCI_write_config8 (Bus, DevFn, CARDBUS_BUS, 1, (int8_t *) &sub_bus) != PCI_SUCCESS) {
			return (-1);
			}
		}

	if (PCI_read_config8 (Bus, DevFn, CARDBUS_SUB_BUS, 1, (int8_t *) &tmp) != PCI_SUCCESS) {
		return (-1);
		}
	if (PCI_write_config8 (Bus, DevFn, CARDBUS_SUB_BUS, 1, (int8_t *) &sub_bus) != PCI_SUCCESS) {
		return (-1);
		}

	PCI_read_config16 (Bus, DevFn, offsetof (struct _pci_config_regs, Status), 1, (char *) &status);
	if (status & PCI_STATUS_CAPABILITIES) {
		if (check_capabilities (Bus, DevFn, Dev))
			return (-1);
		}

	if (scan_windows (Bus, DevFn, Dev) == -1) {
		return (-1);
		}

	if (PCI_read_config8 (Bus, DevFn, offsetof(struct _pci_config_regs, Interrupt_Pin), 1, &tmp) != PCI_SUCCESS) {
		return (-1);
		}
	Dev->IntPin = tmp;
#if	0
#ifdef	__X86__
	if (PCI_read_config8 (Bus, DevFn, offsetof(struct _pci_config_regs, Interrupt_Line), 1, &tmp) != PCI_SUCCESS) {
		return (-1);
		}
	if (tmp == 0) {
		int			num_irq;
		uint32_t	irqlist [SERV_MAX_IRQS];
		num_irq = SERV_MAX_IRQS;
		if ((*pci_drivers->entry->avail_irq) (ll_handle, Bus, DevFn, irqlist, &num_irq) != PCI_SUCCESS) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Unable to allocate CardBus IRQ");
			}
		else {
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "CardBus IRQ %x", irqlist [0]);
			if ((*pci_drivers->entry->map_irq) (ll_handle, Bus, DevFn,
					irqlist [0], Dev->IntPin + 9, PCI_MAP_IRQ) != PCI_SUCCESS) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Unable to map CardBus IRQ");
				}
			else {
				Dev->IrqNumber = irqlist [0];
				tmp = (uint8_t) Dev->IrqNumber;
				if (PCI_write_config8 (Bus, DevFn, offsetof(struct _pci_config_regs, Interrupt_Line), 1, &tmp) != PCI_SUCCESS) {
					return (-1);
					}
				}
			}
		}
#endif
#endif

	PCI_read_config32 (Bus, DevFn, offsetof(struct _pci_cardbus_config_regs, Subsystem_Vendor_Id), 1, (char*)&dword);
	Dev->SubVendorId = dword & 0xFFFF;
	Dev->SubDeviceId = dword >> 16;

	if ((map = find_bus (lastbus)) == NULL) {
		if ((map = alloc_bus (Bus, lastbus)) == NULL)
			return (-1);
		}

	if (Dev->bmap == NULL)		//Link the Device and BUSMAP structures
		Dev->bmap = map;
	if (map->bdev == NULL)
		map->bdev = Dev;
	Dev->Persistent = 1;

	return (0);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	int			traverse_back ()

{
Device		*dev;
BUSMAP		*map;
uint32_t	dword;
int			devfn;
uint8_t		bus, sbus;

	sbus = cur_scan_bus;
	if ((map = find_bus (sbus)) == NULL) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Bus not found!");
		return (-1);
		}
	dev = map->bdev;
	bus = dev->BusNumber;
	devfn = dev->DevFuncNumber;
	cur_scan_bus = map->primary_bus;
	if ((PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
		return (-1);
	dword >>= 8;
	if ((dword & 0xffff00) != (PCI_CLASS_BRIDGE)) {
		PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
		dword = PCI_SECONDARY_BUS_INSERT (dword, sbus);
		dword = PCI_SUBORDINATE_BUS_INSERT (dword, lastbus - 1);
		PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "tb Bus %x - Devfn %x - dword %x", bus, devfn, dword);
		}
	sbus--;
	while (map->prev) {
		map = map->prev;
		dev = map->bdev;
		if (dev == NULL) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "dev is NULL!");
			break;
			}
		if (dev->BusNumber == bus)
			continue;
		bus = dev->BusNumber;
		devfn = dev->DevFuncNumber;
#if	0
		if ((PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
			return (-1);
		dword >>= 8;
		if ((dword & 0xffff00) != (PCI_CLASS_BRIDGE)) {
			PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
			dword = PCI_SUBORDINATE_BUS_INSERT (dword, lastbus - 1);
			PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Primary_Bus_Number), 1, &dword);
			if (verbose > 2)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "tb1 Bus %x - Devfn %x - dword %x", bus, devfn, dword);
			}
#endif
		map->subordinate_bus = lastbus - 1;
		sbus--;
		}
	return (PCI_SUCCESS);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	int		alloc_resource (uint64_t address, uint32_t size, uint32_t flags)

{
rsrc_request_t	rreq;

	if (! address || ! size)
		return (EOK);
	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.start = address;
	rreq.end = address + size;
	rreq.length = size;
	rreq.flags = (flags | RSRCDBMGR_FLAG_RANGE);

	if (rsrcdbmgr_attach (&rreq, 1) == -1) {
		if (verbose > 2) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Unable to seed resource Addr %llx - Size %x", address, size);
			}
		return (ENOMEM);
		}
//	my_slogf (_SLOGC_PCI, _SLOG_INFO, "Alloc resource %llx - Size %x - Flags %x", address, size, flags);
	return (EOK);
}

/**************************************************************************/
/* Power down a PCI device                                                */
/**************************************************************************/

static	int	power_down_device (uint8_t bus, uint32_t devfunc, Device *dev)

{
uint8_t		cap_ptr = dev->PmCap;
uint16_t	cmd, sav_cmd, pwr = dev->PmState;

	if (PCI_read_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
		1, &sav_cmd) != PCI_SUCCESS)
		return (-1);
	cmd = sav_cmd & ~0x7;
	if (PCI_write_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
		1, &cmd) != PCI_SUCCESS)
		return (-1);
	if (PCI_write_config16 (bus, devfunc, cap_ptr + 4, 1, &pwr) != PCI_SUCCESS)
		return (-1);
	if (PCI_write_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
		1, &sav_cmd) != PCI_SUCCESS)
		return (-1);
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Powered down bus %x - devfunc %x", bus, devfunc);
	return (EOK);
}

/**************************************************************************/
/* Power up a PCI device                                                  */
/**************************************************************************/

static	int	power_up_device (uint8_t bus, uint32_t devfunc, Device *dev)

{
uint8_t		cap_ptr = dev->PmCap;
uint16_t	cmd, sav_cmd, pwr = dev->PmState;
uint32_t	pci_regs[20];

	if (pwr & 0x3) {
		if (PCI_read_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
			1, &sav_cmd) != PCI_SUCCESS)
			return (-1);
		cmd = sav_cmd & ~0x7;
		if (PCI_write_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
			1, &cmd) != PCI_SUCCESS)
			return (-1);
		if (PCI_read_config32 (bus, devfunc, 0, 20, pci_regs) != PCI_SUCCESS)
			return (-1);
		pwr &= ~3;
		if (PCI_write_config16 (bus, devfunc, cap_ptr + 4, 1, &pwr) != PCI_SUCCESS)
			return (-1);
		delay (20);
		if (PCI_write_config32 (bus, devfunc, 0, 20, pci_regs) != PCI_SUCCESS)
			return (-1);
		if (PCI_write_config16 (bus, devfunc, offsetof(struct _pci_config_regs, Command),
			1, &sav_cmd) != PCI_SUCCESS)
			return (-1);
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Powered up bus %x - devfunc %x", bus, devfunc);
		}
	return (EOK);
}

/**************************************************************************/
/* Determine the size of memory/io window                                 */
/**************************************************************************/

static uint64_t	window_size (uint8_t BusNumber, uint32_t DevFuncNumber, uint32_t Offset, uint32_t type, uint32_t reg_64)

{
uint32_t	cur_addr [2], tmp1, tmp_addr;
uint64_t	size, tmp, tmp2;

	cur_addr [1] = tmp_addr = 0xffffffff;
	PCI_read_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&cur_addr [0]);
	if (reg_64) {
		PCI_read_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&cur_addr [1]);
		}
	tmp1 = 0xffffffff;
	PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&tmp1);
	if (reg_64) {
		PCI_write_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&tmp1);
		}
	PCI_read_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&tmp1);
	tmp = tmp1;
	if (reg_64) {
		PCI_read_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&tmp_addr);
		tmp2 = tmp_addr;
		tmp |= (tmp2 << 32);
		}
	switch( type ) {
		case PCI_IO_SPACE:
			tmp = PCI_IO_ADDR (tmp);
			break;
		case PCI_MEM_SPACE:
			tmp = PCI_MEM_ADDR (tmp);
			break;
		case PCI_ROM_SPACE:
			tmp = PCI_ROM_ADDR (tmp);
			break;
		}
	if (tmp) {
		if (tmp < 0x10000)
			tmp |= 0xffff0000;
		if (tmp < 0x100000000)
			tmp |= 0xffffffff00000000;
		}

/* Some stupid chips are pre-programmed with a base address, so default */
	if (tmp1 && cur_addr [0] && tmp_addr == cur_addr [1]) {
		tmp_addr = 0;
		PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&tmp_addr);
		PCI_read_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&tmp_addr);
		}
	PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&cur_addr [0]);
	if (reg_64) {
		PCI_write_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&cur_addr [1]);
		}
	if (tmp_addr == tmp1)
		size = DEFAULT_SIZE;
	else
		size = (~tmp) + 1;
	return (size);
}

/**************************************************************************/
/* scan base registers                                                    */
/**************************************************************************/

static int	scan_windows (uint8_t BusNumber, uint32_t DevFuncNumber, Device *device)

{
int			i, reg_64, device_enabled;
int			cardbus_device = 0;
uint32_t	Offset;
uint32_t	dword;
uint64_t	addr64, size;
uint16_t	oldcr;
struct		BaseRegister *BaseReg;

	// disable
	oldcr = off_pci (BusNumber, DevFuncNumber);
	device_enabled = (oldcr & (PCI_COMMAND_IO_ENABLE | PCI_COMMAND_MEM_ENABLE));
	if (! program_display_device) {
		uint32_t	class = device->ClassCode & 0xff0000;

		oldcr |= (class == PCI_CLASS_DISPLAY || class == PCI_CLASS_BRIDGE ||
				class == PCI_CLASS_PROCESSOR) ? PCI_COMMAND_MASTER_ENABLE : 0;
		}

	if (device->ClassCode == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
		cardbus_device = 1;
		}

	// probe
	for (i = 0; i < (cardbus_device ? 1 : 6); i++) {
		BaseReg = &device->BaseReg [i];
		Offset = offsetof(struct _pci_config_regs, Base_Address_Regs [i]);
		PCI_read_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&dword);
		if (PCI_IS_IO (dword)) {
			if (!(size = window_size (BusNumber, DevFuncNumber, Offset, PCI_IO_SPACE, 0))) {
				continue;
				}
			BaseReg->Size = size;
			brqt [BusNumber].io_reqd += size;
			BaseReg->Flags |= PCI_IO_SPACE;
			if (check_range ((uint64_t) (dword & ~3), (uint64_t) size, BaseReg->Flags, BusNumber)) {
				my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Range check failed (IO) - Dev %x - Vend %x - Class %x - Addr %x - Size %llx",
					device->DeviceId, device->VendorId, device->ClassCode, dword & ~3, size);
				if ((device->ClassCode & 0xf0000) != PCI_CLASS_STORAGE &&
					(device->ClassCode & 0xf0000) != PCI_CLASS_DISPLAY) {
					device_enabled = 0;		//Not in valid range
					}
				}
			if (verbose)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "IO %x %llx", PCI_IO_ADDR (dword), size);
			BaseReg->Address = device_enabled ? dword : (dword &= 3);	//Leave reserved bits
			if (! device_enabled || !(dword & ~3)) {
				PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&dword);
				}
			else {
				uint64_t Address;

				(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) BaseReg->Flags | PCI_MAP_ADDR_CPU);

				if (alloc_resource ((uint32_t) PCI_IO_ADDR (Address), (uint32_t) size, RSRCDBMGR_IO_PORT)) {
					my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc failed %llx - Size %llx", Address, size);
					}
				BaseReg->Address = Address;
				BaseReg->Flags |= PCI_DEVICE_ENABLED;
				}
			}
		else {
			reg_64 = 0;
			if ((dword & 0x6) == 0x4) {		//64-bit address
				size = 0;
				PCI_read_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&size);
				addr64 = (size << 32) | dword;
				reg_64 = 1;
				i++;
				}
			else {
				addr64 = (uint64_t) dword;
				}
			if ((size = window_size (BusNumber, DevFuncNumber, Offset, PCI_MEM_SPACE, reg_64)) < 8) {
				continue;
				}
			BaseReg->Size = size;
			BaseReg->Flags |= PCI_MEM_SPACE;
			if (addr64 & 0x08) {		//Test for prefetchable
				BaseReg->Flags |= PCI_MEM_PREFETCH;
				brqt [BusNumber].pref_reqd += size;
				}
			else
				brqt [BusNumber].mem_reqd += size;
			if (check_range ((addr64 & ~0x0f), (uint64_t) size, BaseReg->Flags, BusNumber)) {
				my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Range check failed (MEM) - Dev %x - Vend %x - Class %x - Addr %llx - Size %llx",
					device->DeviceId, device->VendorId, device->ClassCode, addr64 & ~0x0f, size);
				if ((device->ClassCode & 0xf0000) != PCI_CLASS_STORAGE &&
					(device->ClassCode & 0xf0000) != PCI_CLASS_DISPLAY) {
					device_enabled = 0;		//Not in valid range
					}
				}
			if (verbose)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "MEM %llx %llx %s", PCI_MEM_ADDR (addr64), size, addr64 & 0x08 ? "pref" : "");
			BaseReg->Address = device_enabled ? addr64 : (addr64 &= 0x0f);
			if (! device_enabled) {
				PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&addr64);
				if (reg_64) {
					addr64 >>= 32;
					PCI_write_config32 (BusNumber, DevFuncNumber, Offset + 4, 1, (char *)&addr64);
					}
				}
			else {
				uint64_t Address;

				(*pci_drivers->entry->map_addr) (ll_handle, addr64, &Address, (uint32_t) BaseReg->Flags | PCI_MAP_ADDR_CPU);

				if (alloc_resource (PCI_MEM_ADDR (Address), (uint32_t) size, RSRCDBMGR_PCI_MEMORY)) {
					my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc failed %llx - Size %llx", Address, size);
					}
				BaseReg->Address = Address;
				BaseReg->Flags |= PCI_DEVICE_ENABLED;
				}
			}
		}
	// enable
	if (cardbus_device) {
		on_pci (BusNumber, DevFuncNumber, oldcr);
		return (PCI_SUCCESS);
		}

	Offset = offsetof (struct _pci_config_regs, ROM_Base_Address);
	PCI_read_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&dword);
	BaseReg = &device->ROMBase;
	if ((size = window_size (BusNumber, DevFuncNumber, Offset, PCI_MEM_SPACE, 0)) >= 8) {
		BaseReg->Address = device_enabled ? dword : (dword &= 0x7fe);
		BaseReg->Size = size;
		brqt [BusNumber].mem_reqd += size;
		BaseReg->Flags |= PCI_ROM_SPACE;
		if (check_range ((uint64_t) (dword & ~1), (uint64_t) size, BaseReg->Flags, BusNumber)) {
			if (verbose)
				my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Range check failed (ROM) - Dev %x - Vend %x - Class %x - Addr %x - Size %llx",
					device->DeviceId, device->VendorId, device->ClassCode, dword & ~1, size);
			}
		if (verbose)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "ROM %x - %llx", PCI_ROM_ADDR (dword), size);
		}
	if (!device_enabled) {
		dword &= 0x7fe;		//Leave reserved bits
		PCI_write_config32 (BusNumber, DevFuncNumber, Offset, 1, (char *)&dword);
		}
	else {
		uint64_t Address;

		(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) BaseReg->Flags | PCI_MAP_ADDR_CPU);

// This must also be fixed!!!
//		if (alloc_resource (Address, (uint32_t) size, (RSRCDBMGR_MEMORY | RSRCDBMGR_PCI_MEMORY))) {
//			return (-1);
//			}
		alloc_resource (Address, (uint32_t) size, (RSRCDBMGR_MEMORY | RSRCDBMGR_PCI_MEMORY));

		BaseReg->Flags |= PCI_DEVICE_ENABLED;
		on_pci (BusNumber, DevFuncNumber, oldcr);
		if (verbose) {
			for (i = 0; i < 6; i++) {
				BaseReg = &device->BaseReg [i];
				if (BaseReg->Flags) {
					if (BaseReg->Flags & PCI_IO_SPACE)
						my_slogf (_SLOGC_PCI, _SLOG_INFO, "IO %llx %llx", PCI_IO_ADDR (BaseReg->Address), BaseReg->Size);
					else
						if (BaseReg->Flags & PCI_MEM_SPACE)
							my_slogf (_SLOGC_PCI, _SLOG_INFO, "MEM %llx %llx", PCI_MEM_ADDR (BaseReg->Address), BaseReg->Size);
						else
							my_slogf (_SLOGC_PCI, _SLOG_INFO, "ROM %llx %llx", PCI_ROM_ADDR (BaseReg->Address), BaseReg->Size);
					}
				}
			}
		}
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* scan given device, store config                                        */
/**************************************************************************/

static int	scan_device (uint8_t BusNumber, uint8_t SecBus, uint32_t DevFuncNumber)

{
uint32_t 	dword;
uint8_t 	byte;
int			i, index;
uint16_t	vendor, deviceid, status;
Device		*device, *Dev;
pci_bus_t	*pcib;
BUSMAP		*map;

	if (!(Dev = calloc (1, sizeof (*Dev))))
		return (-1);

	device = DeviceHead;
	if (verbose)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "scan_device Bus %d - Device %d - Function %d", BusNumber, DevFuncNumber >> 3, DevFuncNumber & 0x7);
	Dev->BusNumber = BusNumber;
	Dev->DevFuncNumber = DevFuncNumber;
	// vendor, device
	PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Vendor_ID), 1, (char*)&dword);
	Dev->VendorId = vendor = dword & 0xFFFF;
	Dev->DeviceId = deviceid = dword >> 16;

	if (device) {
		index = 0;
		while (device) {
			if (device->VendorId == vendor && device->DeviceId == deviceid) {
				index = device->Index + 1;
				}
			device = device->next;
			}
		}
	else
		index = 0;

	Dev->Index = index;
	// class code
	PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Revision_ID), 1, (char*)&dword);
	Dev->RevisionId = (uchar_t) dword & 0xff;
	dword >>= 8;
	Dev->ClassCode = dword;
	if (verbose)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Vendor %x Device %x Index %d ClassCode %x", vendor, deviceid, index, dword);

	if ((map = find_bus (SecBus)) == NULL) {
		if ((map = alloc_bus (BusNumber, SecBus)) == NULL) {
			free (Dev);
			return (-1);
			}
		}
	pcib = map->pci_bus;
	if (Dev->bmap == NULL)		//Link the Device and BUSMAP structures
		Dev->bmap = map;
	if (map->bdev == NULL)
		map->bdev = Dev;

	if (DeviceHead == NULL) {
		DeviceHead = Dev;
		DeviceTail = Dev;
		}
	else {
		DeviceTail->next = Dev;
		DeviceTail = Dev;
		}
	numdevices++;

	pcib->ndev++;

	// header type
	PCI_read_config8 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Header_Type), 1, (char *) &byte);
	Dev->HeaderType = byte;

	if (Dev->ClassCode == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
		Setup_CardBus_Bridge (BusNumber, DevFuncNumber, Dev);
		}

	// pci status
	PCI_read_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Status), 1, (char *) &status);
	if ((status & PCI_STATUS_BACKTOBACK_OKAY) == 0)
		pcib->fast_b2b = 0;
	if ((status & PCI_STATUS_66MHZ_OKAY) == 0)
		pcib->freq66 = 0;
	i = (status & PCI_STATUS_DEVSEL_MASK) >> PCI_STATUS_DEVSEL_SHIFT;
	if (i > pcib->devsel)
		pcib->devsel = i;

	PCI_read_config8 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Min_Gnt), 1, (char *) &byte);
	Dev->Min_Gnt = byte;
	PCI_read_config8 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Max_Lat), 1, (char *) &byte);
	Dev->Max_Lat = byte;
    if (Dev->Min_Gnt != 0 || Dev->Max_Lat != 0) {
		/* find largest minimum grant time of all devices */
		if (Dev->Min_Gnt != 0 && Dev->Min_Gnt > pcib->Min_Gnt)
			pcib->Min_Gnt = Dev->Min_Gnt;
	
		/* find smallest maximum latency time of all devices */
		if (Dev->Max_Lat != 0 && Dev->Max_Lat < pcib->Max_Lat)
		    pcib->Max_Lat = Dev->Max_Lat;
	
		/* subtract our minimum on-bus time per second from bus bandwidth */
		pcib->bandwidth -= Dev->Min_Gnt * 132000000 / (Dev->Min_Gnt + Dev->Max_Lat);
		if (verbose) {
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Min_Gnt = %d - Max_Lat = %d - Bandwidth = %d", Dev->Min_Gnt, Dev->Max_Lat, pcib->bandwidth);
			}
		}

	// type 0 header only ie. not a pci bridge
	// Stop fetching for PCI_SUBCLASS_BRIDGE_OTHER
	if (((Dev->HeaderType & 0x7F) != 0x00) || ((Dev->ClassCode & 0xff0000) == PCI_CLASS_BRIDGE) &&
			((Dev->ClassCode & 0xff00) != PCI_SUBCLASS_BRIDGE_PCMCIA) &&
			(! enum_xtrans_bridge || ((Dev->ClassCode & 0xff00) != PCI_SUBCLASS_BRIDGE_OTHER)) )
		return (PCI_SUCCESS);

	if (((Dev->ClassCode & 0xff0000) == PCI_CLASS_BRIDGE) &&
		((Dev->ClassCode & 0xff00) == PCI_SUBCLASS_BRIDGE_OTHER)) {
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Forcing enumeration of bridge type OTHER : %d",BusNumber);
		}

	/* Check the capabilities */
	
	PCI_read_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Status), 1, (char *) &status);
	if (status & PCI_STATUS_CAPABILITIES) {
		if (check_capabilities (BusNumber, DevFuncNumber, Dev))
			return (-1);
		}

	/* base registers */
	scan_windows (BusNumber, DevFuncNumber, Dev);

	PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Sub_Vendor_ID), 1, (char*)&dword);
	Dev->SubVendorId = dword & 0xFFFF;
	Dev->SubDeviceId = dword >> 16;

	// irq
	PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Interrupt_Pin), 1, (char *) &byte);
	if (byte == 0x0) {
		Dev->IrqNumber = 0xffffffff;
		Dev->IntPin = 0;
		}
	else {
		Dev->IntPin = byte;
		PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Interrupt_Line), 1, (char *) &byte);
		Dev->IrqNumber = (byte == 0xff) ? 0xffffffff : byte;
#ifndef	__PPC__
		if (byte != 0xff) {
			if (pci_reserve_irq ((uint32_t) byte)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Unable to reserve IRQ %d", byte);
				}
			}
#endif
		}
	PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Latency_Timer), 1, (char*)&byte);
	if (!byte) {		//Plug the latency timer if not initialised
		byte = 0x20;	//Default value
		PCI_write_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Latency_Timer), 1, (char*)&byte);
		}
	return (PCI_SUCCESS);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

static	int		setup_pci2pci_bridges (uint8_t BusNumber)

{
int			DevNumber, DevFuncNumber, NumFuncs;
int			FuncNumber, found = 0;
uint32_t	dword, tmp;
uint8_t		byte;

	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Setup bridge %d - Lastbus = %d", BusNumber, lastbus);
	for (DevNumber = 0; DevNumber < 32; DevNumber++) {
		NumFuncs = 1;
		for (FuncNumber = 0; FuncNumber < NumFuncs; FuncNumber++) {
			DevFuncNumber = (DevNumber << 3) | FuncNumber;
			cur_scan_bus = BusNumber;
			if ((PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
				break;
			if (dword == 0 || dword == 0xffffffff)
				continue;
			dword >>= 8;
			if (FuncNumber == 0) {
				if (PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Header_Type), 1, (char *)&byte) != PCI_SUCCESS)
					break;
				if (byte & 0x80)
					NumFuncs = 8;
				}
			if ((dword & 0xff0000) != (PCI_CLASS_BRIDGE))
				continue;
			if (BusNumber == host_bus && DevFuncNumber == host_devfunc)
				continue;
			tmp = dword & 0xff00;
			if (tmp == PCI_SUBCLASS_BRIDGE_PCI) {
				found = 1;

				configure_bridge (BusNumber, DevFuncNumber);

				if (scan_device (BusNumber, lastbus - 1, DevFuncNumber) != PCI_SUCCESS) {
					return (-1);
					}

				if (! setup_pci2pci_bridges (lastbus - 1)) {
					if (traverse_back ()) {
						return (0);
						}
					}
				found = 0;
				}
			}
		if (lastbus >= scan_buses)
			break;
		}
	return (found);
}

/**************************************************************************/
/* find host bridge                                                       */
/**************************************************************************/

static	int		find_host_bridge (int bus, int devfn)

{
int			status;
uint32_t	dword;

	if ((status = PCI_read_config32 (bus, devfn, offsetof(struct _pci_config_regs, Revision_ID), 1, (char *)&dword)) != PCI_SUCCESS) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config status %x", status);
		return (-1);
		}
	if (dword == 0 || dword == 0xffffffff)
		return (-1);
	dword >>= 8;
	if (host_bridge_class == 0xffffffff) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_INFO, "Found host bridge bus %x - devfn %x - class %x", bus, devfn, dword);
		host_bridge_class = dword;
		if (host_bus == 0xff) {
			host_bus = bus;
			host_devfunc = devfn;
			}
		return (PCI_SUCCESS);
		}
	return (-1);
}

/**************************************************************************/
/* A debug function to scan the pci bus for all devices.                  */
/**************************************************************************/

void	find_all (void)

{
int		bus, dev, func, numfunc, devfunc;
int		vend, device;
uint32_t	dword;
uint8_t		byte;

	for (bus = 0; bus < 0x5; bus++) {
		for (dev = 0; dev < 32; dev++) {
			numfunc = 1;
			for (func = 0; func < numfunc; func++) {
				devfunc = (dev << 3) | func;
				if ((PCI_read_config32 (bus, devfunc, offsetof(struct _pci_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
					break;
				if (dword == 0 || dword == 0xffffffff)
					continue;
				dword >>= 8;
				if (func == 0) {
					if (PCI_read_config8 (bus, devfunc, offsetof(struct _pci_config_regs, Header_Type), 1, (char *)&byte) != PCI_SUCCESS)
						break;
					if (byte & 0x80)
						numfunc = 8;
					}
				PCI_read_config32 (bus, devfunc, offsetof(struct _pci_config_regs, Vendor_ID), 1, (char*)&dword);
				vend = dword & 0xFFFF;
				device = dword >> 16;
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Bus %d - Dev %d - Func %d - Vend %x - Device %x",
					bus, dev, func, vend, device);
				}
			}
		}
}

/**************************************************************************/
/* scan the PCI bus                                                       */
/**************************************************************************/

static int		scan_all (void)

{
int			DevNumber, FuncNumber, DevFuncNumber;
int			NumFuncs, lb;
uint32_t	dword;
uint8_t		byte, BusNumber = 0;
Device		*nextdev, *Device = DeviceHead;

	for (lb = 0; lb < 20; lb++) {
		brqt [lb].io_reqd = 0;
		brqt [lb].mem_reqd = 0;
		brqt [lb].pref_reqd = 0;
		}

	if (DeviceHead) {
		while (nextdev = Device->next)
			free (Device);
		free (nextdev);
		DeviceHead = DeviceTail = NULL;
		}
	numdevices = 0;

	if (verbose > 3) {
		find_all ();
		}

	for (DevNumber = 0; DevNumber < 32; DevNumber++) {
		NumFuncs = 1;
		for (FuncNumber = 0; FuncNumber < NumFuncs; FuncNumber++) {
			DevFuncNumber = (DevNumber << 3) | FuncNumber;
			if (FuncNumber == 0) {
				if (PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Header_Type), 1, (char *)&byte) != PCI_SUCCESS)
					break;
				if (byte & 0x80)
					NumFuncs = 8;
				}
			if (find_host_bridge (BusNumber, DevFuncNumber) == PCI_SUCCESS) {
				if (configure_bridge (BusNumber, DevFuncNumber) != PCI_SUCCESS)		//Configure the host/pci bridge first
					return (-1);

				lb = lastbus ? lastbus - 1 : lastbus;
				if (scan_device (BusNumber, lb, DevFuncNumber) != PCI_SUCCESS) {
					return (-1);
					}
				setup_pci2pci_bridges (lb);
				}
			}
		}


	for (BusNumber = 0; BusNumber < scan_buses; BusNumber++) {
		for (DevNumber = 0; DevNumber < 32; DevNumber++) {
			NumFuncs = 1;
			for (FuncNumber = 0; FuncNumber < NumFuncs; FuncNumber++) {
				DevFuncNumber = (DevNumber << 3) | FuncNumber;
				if ((PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Vendor_ID), 1, (char *)&dword) != PCI_SUCCESS))
					break;
				if (dword == 0 || dword == 0xffffffff)
					continue;
				if (FuncNumber == 0) {
					if (PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Header_Type), 1, (char *)&byte) != PCI_SUCCESS)
						break;
					if (byte & 0x80)
						NumFuncs = 8;
					}
				if ((PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
					break;
				if (dword == 0 || dword == 0xffffffff)
					break;
				dword >>= 8;
				dword &= 0xffff00;
				if (dword == (PCI_CLASS_BRIDGE) || dword == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_PCI)) {
					continue;		//Bridges have already been setup
					}

				if (scan_device (BusNumber, BusNumber, DevFuncNumber) != PCI_SUCCESS) {
					return (-1);
					}
				}
			}
		}
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Rescan the PCI busses for devices that have been inserted/removed.     */
/**************************************************************************/

static	int		rescan (void)

{
typedef struct {
	int			bus;
	int			devfn;
	} RESCAN;
int			DevNumber, FuncNumber, DevFuncNumber;
int			NumFuncs, found, num_dev, count, resptr;
uint32_t	dword;
uint8_t		byte, BusNumber;
Device		*nextdev, *prev = NULL;
RESCAN		*res, *resp;
struct 		BaseRegister *BaseReg;
int			i;

	num_dev = numdevices;
	count = resptr = found = 0;
	if ((res = calloc (100, sizeof (RESCAN))) == NULL)
		return (-1);
	resp = res;
	for (BusNumber = 0; BusNumber <= scan_buses; BusNumber++) {
		for (DevNumber = 0; DevNumber < 32; DevNumber++) {
			NumFuncs = 1;
			for (FuncNumber = 0; FuncNumber < NumFuncs; FuncNumber++) {
				DevFuncNumber = (DevNumber << 3) | FuncNumber;
				if ((PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Vendor_ID), 1, (char *)&dword) != PCI_SUCCESS))
					break;
				if (dword == 0 || dword == 0xffffffff)
					continue;
				if (FuncNumber == 0) {
					if (PCI_read_config8 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Header_Type), 1, (char *)&byte) != PCI_SUCCESS)
						break;
					if (byte & 0x80)
						NumFuncs = 8;
					}
				nextdev = DeviceHead;
				found = 0;
				do {
					if (nextdev->BusNumber == BusNumber && nextdev->DevFuncNumber == DevFuncNumber) {
						resp->bus = BusNumber;
						resp++->devfn = DevFuncNumber;
						resptr++;
						found = 1;
						break;
						}
					} while (nextdev = nextdev->next);
				count++;
				if (found)
					continue;
				if ((PCI_read_config32 (BusNumber, DevFuncNumber, offsetof(struct _pci_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
					break;
				if (dword == 0 || dword == 0xffffffff)
					break;
				dword >>= 8;
				dword &= 0xffff00;
				if (dword == (PCI_CLASS_BRIDGE) || dword == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_PCI)) {
					continue;		//Bridges have already been setup
					}

				if (scan_device (BusNumber, BusNumber, DevFuncNumber) != PCI_SUCCESS) {
					free (res);
					return (-1);
					}
				}
			}
		}
	if (count < num_dev) {		//Something has been removed!
		numdevices = count;
		nextdev = DeviceHead;
		do {
			for (count = found = 0, resp = res; count < resptr; count++, resp++) {
				if (nextdev->BusNumber == resp->bus && nextdev->DevFuncNumber == resp->devfn) {
					found = 1;
					break;
					}
				}
			if (! found) {
				if (prev == NULL) {
					DeviceHead = nextdev->next;
					}
				else {
					prev->next = nextdev->next;
					}
				for (i = 0; i < 6; i++) {	//Free the resources
					BaseReg = &nextdev->BaseReg [i];
					if (BaseReg->Address)
						BaseReg->Address &= PCI_IO_ADDR (BaseReg->Address) ? 0x3 : 0x0f;
					}
				BaseReg = &nextdev->ROMBase;
				if (BaseReg->Address)
					BaseReg->Address &= 0x7fe;

				// Don't free CardBus interrupts
				if (nextdev->bmap->bdev->ClassCode != (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
					if (nextdev->IrqNumber != 0 && nextdev->IrqNumber != 0xffffffff) {
						pci_free_irq (nextdev->IrqNumber);
						}
					}
				if (nextdev->io_alloc.Address) {
					pci_free_resource (&nextdev->io_alloc, nextdev->BusNumber);
					}
				if (nextdev->mem_alloc.Address) {
					pci_free_resource (&nextdev->mem_alloc, nextdev->BusNumber);
					}
				if (nextdev->pref_alloc.Address) {
					pci_free_resource (&nextdev->pref_alloc, nextdev->BusNumber);
					}
				if (nextdev->next == NULL) {	//Last item on queue
					DeviceTail = prev;
					free (nextdev);
					break;
					}
				else {
					free (nextdev);
					nextdev = prev;
					continue;
					}
				}
			prev = nextdev;
			} while (nextdev = nextdev->next);
		}
	free (res);
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Search through the Device list looking for a particular class of       */
/* device at a specific index.                                            */
/**************************************************************************/

static	Device	*search_class (uint32_t class, uint32_t idx)

{
Device 		*device = DeviceHead;
int			oidx = 0;

	pthread_mutex_lock (&serv_mutex);
	while (device) {
		if (device->ClassCode == class) {
			if (oidx == idx) {
				pthread_mutex_unlock (&serv_mutex);
				return (device);
				}
			else {
				oidx++;
				}
			}
		device = device->next;
		}
	pthread_mutex_unlock (&serv_mutex);
	return (NULL);
}

/**************************************************************************/
/* Search through the Device list looking for a particular vendor and     */
/* device id at a specific index.                                         */
/**************************************************************************/

static	Device	*search_vendor (uint32_t vendor, uint32_t vendor_mask, uint32_t idx)

{
Device 		*dev_ptr = DeviceHead;

	pthread_mutex_lock (&serv_mutex);
	while (dev_ptr) {
		if ((dev_ptr->VendorId & vendor_mask) == vendor && dev_ptr->Index == idx) {
			pthread_mutex_unlock (&serv_mutex);
			return (dev_ptr);
			}
		dev_ptr = dev_ptr->next;
		}
	pthread_mutex_unlock (&serv_mutex);
	return (NULL);
}

/**************************************************************************/
/* Search through the Device list looking for a particular vendor and     */
/* device id at a specific index.                                         */
/**************************************************************************/

static	Device	*search_venddev (uint32_t vendor, uint32_t device, uint32_t vendor_mask, uint32_t device_mask, uint32_t idx)

{
Device 		*dev_ptr = DeviceHead;

	pthread_mutex_lock (&serv_mutex);
	while (dev_ptr) {
		if ((dev_ptr->VendorId & vendor_mask) == vendor && (dev_ptr->DeviceId & device_mask) == device && dev_ptr->Index == idx) {
			pthread_mutex_unlock (&serv_mutex);
			return (dev_ptr);
			}
		dev_ptr = dev_ptr->next;
		}
	pthread_mutex_unlock (&serv_mutex);
	return (NULL);
}

/**************************************************************************/
/* Search through the Device list looking for a particular bus number     */
/* and devfuncnumber.                                                     */
/**************************************************************************/

static	Device	*search_busdev (uint8_t bus, uint32_t devfunc)

{
Device 		*device = DeviceHead;

	pthread_mutex_lock (&serv_mutex);
	while (device) {
		if (device->BusNumber == bus && device->DevFuncNumber == devfunc) {
			pthread_mutex_unlock (&serv_mutex);
			return (device);
			}
		device = device->next;
		}
	pthread_mutex_unlock (&serv_mutex);
	return (NULL);
}

/**************************************************************************/
/* Return the number of devices found by a previous scan of the bus.      */
/**************************************************************************/

/* configuration calls */


static	int		PCI_GetVersion (uint16_t *Version, uint16_t *NumSysDevices)

{
	*Version = version;
	*NumSysDevices = numdevices;
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Search through the Device list looking for a device that matches       */
/* DeviceId, VendorId as well as Index. If the device is found, return    */
/* the busnumber and devfuncnumber.                                       */
/**************************************************************************/

static	int		PCI_find_device (uint32_t DeviceId, uint32_t VendorId, uint32_t Index,
							uint8_t *BusNumber, uint32_t *DevFuncNumber)

{
Device 		*device;

	if ((device = search_venddev (VendorId, DeviceId, 0xffff, 0xffff, Index)) == NULL) {
		return (PCI_DEVICE_NOT_FOUND);
		}
	*BusNumber = device->BusNumber;
	*DevFuncNumber = device->DevFuncNumber;
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Search the Device list looking for a class of device (ClassCode)       */
/* at a specific Index. If found, return the busnumber and devfuncnumber. */
/**************************************************************************/

static	int		PCI_find_class (uint32_t ClassCode, uint32_t Index, uint8_t *BusNumber,
							uint32_t *DevFuncNumber)

{
Device 		*device;

	if (BusNumber == NULL || DevFuncNumber == NULL) {
		return (PCI_UNSUPPORTED_FUNCT);
		}
	if ((device = search_class (ClassCode, Index)) == NULL) {
		return (PCI_DEVICE_NOT_FOUND);
		}
	*BusNumber = device->BusNumber;
	*DevFuncNumber = device->DevFuncNumber;
	return (PCI_SUCCESS);
}

/**************************************************************************/
/* This routine checks to see whether an IRQ can be programmed.           */
/* The code first saves the old irq and then writes another irq and reads */
/* it back. If the source and destination irqs agree, then the irq is     */
/* programmable.                                                          */
/**************************************************************************/

int		irq_ok (uint8_t bus, uint32_t devfn)

{
uint8_t		old_irq, tmp [2];

	if (PCI_read_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Line), 1, &old_irq) != PCI_SUCCESS) {
		return (0);
		}
	if (old_irq != 0xff && old_irq < 16) {
		return (1);					//IRQ is OK
		}
	tmp [0] = 0x09;
	if (PCI_write_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Line), 1, &tmp [0]) != PCI_SUCCESS) {
		return (0);
		}
	if (PCI_read_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Line), 1, &tmp [1]) != PCI_SUCCESS) {
		return (0);
		}
	if (PCI_write_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Line), 1, &old_irq) != PCI_SUCCESS) {
		return (0);
		}
	return ((tmp [0] == tmp [1]) ? 1 : 0);	//IRQ is OK if tmp [0] == tmp [1]
}

/**************************************************************************/
/* This routine checks to see whether the interrupt pin of a non-zero     */
/* function is the same as function zero.                                 */
/**************************************************************************/

int		same_int (uint8_t bus, uint8_t devfn)

{
uint8_t		int_pin, int_pin1;

	if (PCI_read_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Pin), 1, &int_pin) != PCI_SUCCESS) {
		return (0);
		}
	if (PCI_read_config8 (bus, (devfn & 0xf8), offsetof (struct _pci_config_regs, Interrupt_Pin), 1, &int_pin1) != PCI_SUCCESS) {
		return (0);
		}
	return ((int_pin == int_pin1) ? 1 : 0);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

int		is_not_bridge (uint8_t bus, uint8_t devfn)

{
uint32_t	dword;

	if ((PCI_read_config32 (bus, (devfn & 0xf8), offsetof (struct _pci_bridge_config_regs, Revision_ID), 1, (char *)&dword) != PCI_SUCCESS))
		return (-1);
	dword >>= 8;
	if ((dword & 0xff0000) == (PCI_CLASS_BRIDGE)) {
		return (0);
		}
	return (1);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

int		PCI_translation( void *ll_handle, struct pci_dev_info *ConfigBuffer )

{
uint64_t	iaddr, oaddr;

	iaddr = 0;
	if ((*pci_drivers->entry->map_addr)(ll_handle, iaddr, &oaddr, PCI_IO_SPACE | PCI_MAP_ADDR_PCI) != EOK)
		my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Translate I/O failed");
	else
		ConfigBuffer->CpuIoTranslation = -oaddr;
	if ((*pci_drivers->entry->map_addr)(ll_handle, iaddr, &oaddr, PCI_MEM_SPACE | PCI_MAP_ADDR_PCI) != EOK)
		my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Translate Mem failed");
	else
		ConfigBuffer->CpuMemTranslation = -oaddr;
	if ((*pci_drivers->entry->map_addr)(ll_handle, iaddr, &oaddr, PCI_ISA_SPACE | PCI_MAP_ADDR_PCI) != EOK)
		my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Translate ISA failed");
	else
		ConfigBuffer->CpuIsaTranslation = -oaddr;
	if ((*pci_drivers->entry->map_addr)(ll_handle, iaddr, &oaddr, PCI_BMSTR_SPACE | PCI_MAP_ADDR_PCI) != EOK)
		my_slogf (_SLOGC_PCI, _SLOG_WARNING, "Translate Bmstr failed");
	else
		ConfigBuffer->CpuBmstrTranslation = oaddr;

	return( 0 );
}

/**************************************************************************/
/* Validate the handle                                                    */
/**************************************************************************/

static	Device	*validate_handle (Device *handle)

{
Device 		*dev_ptr = DeviceHead;

	while (dev_ptr) {
		if (dev_ptr == handle)
			return (handle);
		dev_ptr = dev_ptr->next;
		}
	return (NULL);
}

/**************************************************************************/
/* Fill base register config information                                  */
/**************************************************************************/

static	void	fill_base (struct pci_dev_info *ConfigBuffer, Device *dev_ptr)

{
struct		BaseRegister *BaseReg;
int			i;
uint64_t	Address;

	for (i = 0; i < 6; i++) {
		BaseReg = &dev_ptr->BaseReg [i];
		if (! BaseReg->Size)
			continue;
		(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) (BaseReg->Flags | PCI_MAP_ADDR_PCI));
		ConfigBuffer->PciBaseAddress [i] = Address;
		ConfigBuffer->CpuBaseAddress [i] = BaseReg->Address;
		ConfigBuffer->BaseAddressSize [i] = BaseReg->Size;
		}
	BaseReg = &dev_ptr->ROMBase;
	if (BaseReg->Flags & PCI_ROM_SPACE) {
		(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) (BaseReg->Flags | PCI_MAP_ADDR_PCI));
		ConfigBuffer->PciRom = Address;
		ConfigBuffer->CpuRom = BaseReg->Address;
		ConfigBuffer->RomSize = (uint32_t) BaseReg->Size;
		}
	PCI_translation (ll_handle, ConfigBuffer);
}

/**************************************************************************/
/* Fill basic config information                                          */
/**************************************************************************/

static	void	fill_config (struct pci_dev_info *ConfigBuffer, Device *dev_ptr)

{
	ConfigBuffer->DeviceId = dev_ptr->DeviceId;
	ConfigBuffer->VendorId = dev_ptr->VendorId;
	ConfigBuffer->Class = dev_ptr->ClassCode;
	ConfigBuffer->BusNumber = dev_ptr->BusNumber;
	ConfigBuffer->DevFunc = dev_ptr->DevFuncNumber;
	ConfigBuffer->Revision = dev_ptr->RevisionId;
	ConfigBuffer->SubsystemId = dev_ptr->SubDeviceId;
	ConfigBuffer->SubsystemVendorId = dev_ptr->SubVendorId;
	ConfigBuffer->Irq = dev_ptr->IrqNumber;
}

/**************************************************************************/
/* Make a check of the attach flags                                       */
/**************************************************************************/

static	int		check_flags (Device *dev_ptr, uint32_t flags)

{
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "AttachFlags %x - flags %x", dev_ptr->AttachFlags, flags);
	if ((dev_ptr->AttachFlags & PCI_HW_INIT) == 0)
		return ((flags & PCI_HW_INIT) ? 1 : 0);
	if ((dev_ptr->AttachFlags ^ flags) & PCI_HW_INIT) {
		if (hw_override) {
			return (-1);
			}
		return (1);
		}
	return (0);
}

/**************************************************************************/
/* get config of logical device                                           */
/**************************************************************************/

static	Device	*PCI_attach_device (void *handle, uint32_t flags, uint16_t idx,
						struct pci_dev_info *ConfigBuffer, DEV_LIST *list,
						int *local_errno)

{
Device 		*dev_ptr;
struct		BaseRegister *BaseReg, alloc_io_reg, alloc_mem_reg;
uint64_t	Address;
uint32_t	ll_Address;
uint16_t	vendor, device, pci_flg;
uint16_t	vendor_mask, device_mask;
int			i, status, device_enabled = 0;
int			show_pci = 0, next_attach = 0;
int			cardbus_device = 0, re_attach = 0;
uint32_t	flg;
uint32_t	irqlist [SERV_MAX_IRQS];
uint8_t		BusNumber, bus;
uint32_t	DevFuncNumber, devfn;
uint32_t	io_size, mem_size, align, pref_align;
uint32_t	pref_size;
uint64_t	pref_addr, mem_addr, io_addr;
int			mgr_attach;
int			reg_idx = 0;
struct		{
			uint64_t	size;
			int			reg;
			} reg_sort [7];

	*local_errno = EOK;
	mgr_attach = (flags & PCI_MGR_ATTACH) ? 1 : 0;
	if (! (flags & (PCI_INIT_ALL | PCI_INIT_ROM))) {	//Special case for show_pci. Display all information but don't enable device
		show_pci = 1;
		}

	if (handle) {
		if ((dev_ptr = validate_handle ((Device *) handle)) == NULL) {
			if (verbose)
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Invalid handle");
			*local_errno = EINVAL;
			return (NULL);
			}
		if ((status = check_flags (dev_ptr, flags)) == -1) {
			if (verbose > 2)
				my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Path1");
			*local_errno = EINVAL;
			return (NULL);
			}
		else {
			if (status) {
				if (verbose > 2)
					my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Path2");
				pci_flg = off_pci (dev_ptr->BusNumber, dev_ptr->DevFuncNumber);
				device_enabled = re_attach = 1;
				}
			else {
				if (verbose > 2)
					my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Path3");
				fill_config (ConfigBuffer, dev_ptr);
				fill_base (ConfigBuffer, dev_ptr);
				dev_ptr->LockFlg++;			/* set device as locked */
				list->lockflg++;
				for (i = 0; i < NUM_DEVICES; i++) {
					if (list->dev_list [i] == dev_ptr) {
						break;
						}
					if (list->dev_list [i] == NULL) {
						list->dev_list [i] = dev_ptr;
						list->num_devices++;
						break;
						}
					}
				return (dev_ptr);
				}
			}
		}
	else {
		device = ConfigBuffer->DeviceId;
		vendor = ConfigBuffer->VendorId;

		switch (flags & PCI_SEARCH_MASK) {
			case	PCI_SEARCH_VEND:
				vendor_mask = 0xffff;
				if (flags & PCI_MASK_VENDDEV) {
					vendor_mask = ConfigBuffer->SubsystemVendorId;
					}
				dev_ptr = search_vendor (vendor, vendor_mask, idx);
				break;
			case	PCI_SEARCH_VENDEV:
				vendor_mask = device_mask = 0xffff;
				if (flags & PCI_MASK_VENDDEV) {
					vendor_mask = ConfigBuffer->SubsystemVendorId;
					device_mask = ConfigBuffer->SubsystemId;
					}
				dev_ptr = search_venddev (vendor, device, vendor_mask, device_mask, idx);
				break;
			case	PCI_SEARCH_CLASS:
				dev_ptr = search_class (ConfigBuffer->Class, idx);
				break;
			case	PCI_SEARCH_BUSDEV:
				dev_ptr = search_busdev (ConfigBuffer->BusNumber, ConfigBuffer->DevFunc);
				break;
			default:
				if (device || vendor) {
					vendor_mask = device_mask = 0xffff;
					if (flags & PCI_MASK_VENDDEV) {
						vendor_mask = ConfigBuffer->SubsystemVendorId;
						device_mask = ConfigBuffer->SubsystemId;
						}
					dev_ptr = search_venddev (vendor, device, vendor_mask, device_mask, idx);
					}
				else {
					if (ConfigBuffer->Class) {
						dev_ptr = search_class (ConfigBuffer->Class, idx);
						}
					else {
						dev_ptr = search_busdev (ConfigBuffer->BusNumber, ConfigBuffer->DevFunc);
						}
					}
				}

		if (dev_ptr == NULL) {
			*local_errno = ENODEV;
			if (verbose > 2)
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "No device");
			return (NULL);
			}
		else {
			if (dev_ptr->LockFlg) {
				if (!mgr_attach && !show_pci) {
					if ((dev_ptr->AttachFlags & PCI_SHARE)) {
						if (!(flags & PCI_SHARE)) {
							*local_errno = EBUSY;
							}
						else {
							next_attach = 1;
							}
						}
					else {
						*local_errno = EBUSY;
						}
					}
				else {
					next_attach = 1;
					}
					
				}
			}
		}
	if (*local_errno != EOK) {
		if (verbose > 2)
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Busy");
		return (NULL);
		}

	if (! dev_ptr->MgrAttach) {
		dev_ptr->MgrAttach = (uchar_t) mgr_attach;
		}
	else {
		next_attach = 1;
		}

	BusNumber = dev_ptr->BusNumber;
	DevFuncNumber = dev_ptr->DevFuncNumber;
	fill_config (ConfigBuffer, dev_ptr);

	if (dev_ptr->ClassCode == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
		cardbus_device = 1;
		}

	if (show_pci) {
		flags = PCI_INIT_ALL | PCI_INIT_ROM;
		}
	else {
		if (dev_ptr->PmState & 0x03) {
			if (power_up_device (BusNumber, DevFuncNumber, dev_ptr)) {
				*local_errno = EIO;
				return (NULL);
				}
			}
		}

// First add up the size of all active registers, so that one
// chunk of IO and/or Memory space can be allocated.

	io_size = mem_size = pref_size = align = 0;
	mem_addr = io_addr = pref_addr = 0;
	pref_align = 0;
	for (i = 0, flg = PCI_INIT_BASE0; i < (cardbus_device ? 1 : 6); i++, flg <<= 1) {
		if (!(flags & flg))				 //Test user enable flags
			continue;
		BaseReg = &dev_ptr->BaseReg [i];
		if (!BaseReg->Flags)
			continue;
		reg_sort [reg_idx].size = BaseReg->Size;
		reg_sort [reg_idx++].reg = i;
		device_enabled |= (BaseReg->Flags & PCI_DEVICE_ENABLED);
		if (BaseReg->Address < ((BaseReg->Flags & PCI_IO_SPACE) ? 4 : 16))
			device_enabled = 0;			//Temporary fix!!!!!
		if (! device_enabled && ! show_pci && ! next_attach) {
			if (BaseReg->Flags & PCI_IO_SPACE) {
				io_size += ((BaseReg->Size < 256) ? 256 : BaseReg->Size);
				}
			else {		//Must be memory
				if (BaseReg->Flags & PCI_MEM_PREFETCH) {
					pref_size += BaseReg->Size;
					if (BaseReg->Size > pref_align)
						pref_align = BaseReg->Size;
					}
				else {
					if (BaseReg->Size > align) {
						align = BaseReg->Size;
						}
					mem_size += BaseReg->Size;
					}
				}
			}
		}
	if (flags & PCI_INIT_ROM) {
		BaseReg = &dev_ptr->ROMBase;
		if (BaseReg->Flags & PCI_ROM_SPACE && ! cardbus_device) {
			int		d1 = (BaseReg->Flags & PCI_DEVICE_ENABLED);
			if (BaseReg->Address < 2)
				d1 = 0;
			if (! d1 && ! show_pci && ! next_attach) {
				if (BaseReg->Size > align) {
					align = BaseReg->Size;
					}
				mem_size += BaseReg->Size;
				reg_sort [reg_idx].size = BaseReg->Size;
				reg_sort [reg_idx++].reg = 6;
				}
			}
		}

	if (io_size) {
		memset ((char *) &alloc_io_reg, 0, sizeof (struct BaseRegister));
		alloc_io_reg.Flags = PCI_IO_SPACE;
		alloc_io_reg.Size = io_size;
		if ((*local_errno = pci_alloc_resource (&alloc_io_reg, 0, BusNumber)) != PCI_SUCCESS) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc I/O failed");
			return (NULL);
			}
		io_addr = alloc_io_reg.Address;
		dev_ptr->io_alloc.Address = io_addr;
		dev_ptr->io_alloc.Size = io_size;
		dev_ptr->io_alloc.Flags = PCI_IO_SPACE;
		}

	if (mem_size) {
		memset ((char *) &alloc_mem_reg, 0, sizeof (struct BaseRegister));
		alloc_mem_reg.Flags = PCI_MEM_SPACE;
		alloc_mem_reg.Size = mem_size;
		if ((*local_errno = pci_alloc_resource (&alloc_mem_reg, align, BusNumber)) != PCI_SUCCESS) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc mem failed");
			return (NULL);
			}
		mem_addr = alloc_mem_reg.Address;
		dev_ptr->mem_alloc.Address = mem_addr;
		dev_ptr->mem_alloc.Size = mem_size;
		dev_ptr->mem_alloc.Flags = PCI_MEM_SPACE;
		}

	if (pref_size) {
		memset ((char *) &alloc_mem_reg, 0, sizeof (struct BaseRegister));
		alloc_mem_reg.Flags = PCI_MEM_PREFETCH;
		alloc_mem_reg.Size = pref_size;
		if ((*local_errno = pci_alloc_resource (&alloc_mem_reg, pref_align, BusNumber)) != PCI_SUCCESS) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc pref failed");
			return (NULL);
			}
		pref_addr = alloc_mem_reg.Address;
		dev_ptr->pref_alloc.Address = pref_addr;
		dev_ptr->pref_alloc.Size = pref_size;
		dev_ptr->pref_alloc.Flags = PCI_MEM_PREFETCH;
		}

	//Sort the BaseRegs in descending order	
	while (1) {
		int			treg, swap = 0;
		uint64_t	tsize;
		for (i = 0; i < reg_idx - 1; i++) {
			if (reg_sort [i].size < reg_sort [i + 1].size) {
				treg = reg_sort [i].reg;
				tsize = reg_sort [i].size;
				reg_sort [i].size = reg_sort [i + 1].size;
				reg_sort [i].reg = reg_sort [i + 1].reg;
				reg_sort [i + 1].size = tsize;
				reg_sort [i + 1].reg = treg;
				swap = 1;
				}
			}
		if (! swap)
			break;
		}


	if ((mem_size || io_size || pref_size) && ! show_pci) {
		for (i = 0; i < reg_idx; i++) {
			int		reg = reg_sort [i].reg;
			if (reg == 6)
				BaseReg = &dev_ptr ->ROMBase;
			else
				BaseReg = &dev_ptr->BaseReg [reg];
			if (verbose)
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Reg %d - Size %x", reg, BaseReg->Size);
			if (BaseReg->Address < ((BaseReg->Flags & PCI_IO_SPACE) ? 4 : 16) && BaseReg->Size) {
				if (BaseReg->Flags & PCI_IO_SPACE) {
					BaseReg->Address |= io_addr;	//Keep flags
					if (cardbus_device)
						BaseReg->Flags |= PCI_DEVICE_ENABLED;
					io_addr += (BaseReg->Size < 256) ? 256 : BaseReg->Size;
					io_size -= (BaseReg->Size < 256) ? 256 : BaseReg->Size;
					}
				else {
					if (BaseReg->Flags & PCI_MEM_PREFETCH) {
						BaseReg->Address |= pref_addr;	//Keep flags
						if (cardbus_device)
							BaseReg->Flags |= PCI_DEVICE_ENABLED;
						pref_addr += BaseReg->Size;
						pref_size -= BaseReg->Size;
						}
					else {
						BaseReg->Address |= mem_addr;	//Keep flags
						if (cardbus_device)
							BaseReg->Flags |= PCI_DEVICE_ENABLED;
						mem_addr += BaseReg->Size;
						mem_size -= BaseReg->Size;
						}
					}
				}
			}
		}

	for (i = 0, flg = PCI_INIT_BASE0, pci_flg = 0; i < (cardbus_device ? 1 : 6); i++, flg <<= 1) {
		if (!(flags & flg))				 //Test user enable flags
			continue;
		BaseReg = &dev_ptr->BaseReg [i];
		if (! BaseReg->Size)
			continue;
		(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) (BaseReg->Flags | PCI_MAP_ADDR_PCI));
		if (BaseReg->Flags & PCI_IO_SPACE) {
			ConfigBuffer->PciBaseAddress [i] = Address;
			ConfigBuffer->CpuBaseAddress [i] = BaseReg->Address;
			ConfigBuffer->BaseAddressSize [i] = (uint32_t) BaseReg->Size;
			pci_flg |= PCI_ISPACEEN;
			}
		else if (BaseReg->Flags & PCI_MEM_SPACE) {
			ConfigBuffer->PciBaseAddress [i] = Address;
			ConfigBuffer->CpuBaseAddress [i] = BaseReg->Address;
			ConfigBuffer->BaseAddressSize [i] = (uint32_t) BaseReg->Size;
			pci_flg |= PCI_MSPACEEN;
			}
		if (! show_pci) {
			ll_Address = (uint32_t) Address;
			if (PCI_write_config32 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Base_Address_Regs [i]), 1, &ll_Address) != PCI_SUCCESS) {
				*local_errno = ENXIO;
				if (verbose > 2)
					my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Write failed %d", i);
				return (NULL);
				}
			}
		}
	BaseReg = &dev_ptr->ROMBase;
	if ((BaseReg->Flags & PCI_ROM_SPACE) && (flags & PCI_INIT_ROM) && !cardbus_device) {
		(*pci_drivers->entry->map_addr) (ll_handle, BaseReg->Address, &Address, (uint32_t) (BaseReg->Flags | PCI_MAP_ADDR_PCI));
		ConfigBuffer->PciRom = Address;
		ConfigBuffer->CpuRom = BaseReg->Address;
		ConfigBuffer->RomSize = (uint32_t) BaseReg->Size;
		if (! show_pci) {
			ll_Address = ((uint32_t) Address | 1);	//Set address decode enable
			if (PCI_write_config32 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, ROM_Base_Address), 1, &ll_Address) != PCI_SUCCESS) {
				*local_errno = ENXIO;
				if (verbose > 2)
					my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Write failed ROM");
				return (NULL);
				}
			}
		pci_flg |= PCI_MSPACEEN;
		}

	if (show_pci || next_attach) {
		ConfigBuffer->Irq = dev_ptr->IrqNumber;
		}
	else {
		int		num_irq = 0;

		ConfigBuffer->Irq = 0xff;
		if (dev_ptr->IntPin && (flags & PCI_INIT_IRQ) && irq_ok (BusNumber, DevFuncNumber)) {
			bus = BusNumber;
			devfn = DevFuncNumber;
			if (devfn & 0x07 && dev_ptr->IntPin == 1 && is_not_bridge (bus, devfn)
				&& same_int (bus, devfn)) {	//Functions on same IntPin
			/* Changed this from a read_config to avail_irq, as interrupts on */
			/* some boards are larger than 8-bits. */
				num_irq = SERV_MAX_IRQS;
				if ((*pci_drivers->entry->avail_irq) (ll_handle, bus, (devfn & 0xf8), irqlist, &num_irq) != PCI_SUCCESS) {
					*local_errno = ENXIO;
					if (verbose > 2)
						my_slogf (_SLOGC_PCI, _SLOG_ERROR, "No irq");
					return (NULL);
					}
				dev_ptr->IrqNumber = irqlist [0];
				}
			else {
				if (dev_ptr->bmap && ! cardbus_device) {
					// For CardBus devices, the IRQ has to be programmed into the CardBus controller, not the card.
					if (dev_ptr->bmap->bdev->ClassCode == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
						bus = dev_ptr->bmap->bdev->BusNumber;
						devfn = dev_ptr->bmap->bdev->DevFuncNumber;
						cardbus_device = 2;
						}
					}
				num_irq = SERV_MAX_IRQS;
				if ((*pci_drivers->entry->avail_irq) (ll_handle, bus, devfn, irqlist, &num_irq) != PCI_SUCCESS) {
					*local_errno = ENXIO;
					if (verbose > 2)
						my_slogf (_SLOGC_PCI, _SLOG_ERROR, "No irq2");
					return (NULL);
					}
				if (cardbus_device == 2 && dev_ptr->bmap->bdev->IrqNumber) {
					dev_ptr->IrqNumber = dev_ptr->bmap->bdev->IrqNumber;
					}
				else {
					if (! next_attach) {
						if (! (dev_ptr->IrqNumber = pci_alloc_irq (irqlist, num_irq, bus))) {
							if (num_irq != 1) {		//If the allocate failed on a list, big problem!
								*local_errno = ENXIO;
								if (verbose > 2)
									my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Alloc IRQ failed");
								return (NULL);
								}
							else {
								dev_ptr->IrqNumber = irqlist [0];
								}
							}
						else {
							if (cardbus_device == 2) {
								dev_ptr->bmap->bdev->IrqNumber = dev_ptr->IrqNumber;
								}
							}
						}
					}
				}
			ConfigBuffer->Irq = dev_ptr->IrqNumber;
#ifdef	__X86__
			if (map_irqs && dev_ptr->IrqNumber != 0xffffffff && num_irq > 1) {
#else
			if (map_irqs && dev_ptr->IrqNumber != 0xffffffff) {
#endif
				if ((status = (*pci_drivers->entry->map_irq) (ll_handle, bus, devfn,
					dev_ptr->IrqNumber, dev_ptr->IntPin + 9, PCI_MAP_IRQ)) != PCI_SUCCESS) {
					if (verbose > 2)
						my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Map IRQ %x failed %x!", dev_ptr->IrqNumber, status);
					}
				}
			{
			uint8_t		irq = (uint8_t) dev_ptr->IrqNumber;
			if (PCI_write_config8 (bus, devfn, offsetof (struct _pci_config_regs, Interrupt_Line), 1, &irq) != PCI_SUCCESS) {
				*local_errno = ENXIO;
				if (verbose > 2)
					my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Write IRQ failed");
				return (NULL);
				}
			}
			}
		}

	PCI_translation( ll_handle, ConfigBuffer );

	if (flags & PCI_MASTER_ENABLE)
		pci_flg |= PCI_MASTEREN;

	if ((dev_ptr->ClassCode & 0xff0000) == PCI_CLASS_DISPLAY) {
		uint16_t	cr;
		PCI_read_config16 (BusNumber, DevFuncNumber, offsetof (struct _pci_config_regs, Command), 1, (char *)&cr);
		pci_flg |= cr;
		pci_flg |= PCI_COMMAND_IO_ENABLE;		//Always enable I/O on video adapters
		}

	if (! show_pci) {
		on_pci (BusNumber, DevFuncNumber, pci_flg);
		if (! dev_ptr->AttachFlags) {
			dev_ptr->AttachFlags |= flags;	/*Save the flag settings */
			if (! mgr_attach && ! re_attach) {
				dev_ptr->LockFlg++;			/* set device as locked */
				list->lockflg++;
				}
			if (!dev_ptr->Persistent) {
				if (flags & PCI_PERSIST) {
					dev_ptr->Persistent = 1;
					for (i = 0; i < 6; i++) {
						BaseReg = &dev_ptr->BaseReg [i];
						if (BaseReg->Flags)
							BaseReg->Flags |= PCI_DEVICE_ENABLED;
						}
					BaseReg = &dev_ptr->ROMBase;
					if (BaseReg->Flags)
						BaseReg->Flags |= PCI_DEVICE_ENABLED;
					}
				}
			}
		else {
			dev_ptr->AttachFlags |= flags;
			dev_ptr->LockFlg++;			/* set device as locked */
			list->lockflg++;
			}
		if (! list->num_devices) {
			list->dev_list [0] = dev_ptr;
			}
		else {
			for (i = 0; i < NUM_DEVICES; i++) {
				if (list->dev_list [i] == dev_ptr) {
					break;
					}
				if (list->dev_list [i] == NULL) {
					list->dev_list [i] = dev_ptr;
					break;
					}
				}
			}
		list->num_devices++;
		}
	else {		// I really don't think this is necessary!
		if (! re_attach) {
			dev_ptr->LockFlg++;			/* set device as locked */
			list->lockflg++;
			}
		if (handle)
			on_pci (BusNumber, DevFuncNumber, pci_flg);
		}

	return (dev_ptr);
}

/**************************************************************************/
/* Release all the resources associated with a particular device.         */
/**************************************************************************/

static	int		release_resources (Device *device)

{
struct 		BaseRegister *BaseReg;
uint32_t	address = 0;
int			i, enabled = 0;

/* This next line was commented out to fix a problem with the MyCable box */
//	enabled = ((device->ClassCode & 0xff0000) == PCI_CLASS_DISPLAY) ? 1 : 0;
	for (i = 0; i < 6; i++) {
		BaseReg = &device->BaseReg [i];
		enabled |= BaseReg->Flags & PCI_DEVICE_ENABLED;
		if (BaseReg->Address && !enabled) {
			PCI_write_config32 (device->BusNumber, device->DevFuncNumber,
				offsetof (struct _pci_config_regs, Base_Address_Regs [i]), 1, (char *)&address);
			BaseReg->Address &= PCI_IS_IO (BaseReg->Address) ? 0x3 : 0x0f;
			}
		}
	BaseReg = &device->ROMBase;
	enabled |= BaseReg->Flags & PCI_DEVICE_ENABLED;
	if (BaseReg->Address && !(BaseReg->Flags & PCI_DEVICE_ENABLED)) {
		PCI_write_config32 (device->BusNumber, device->DevFuncNumber,
			offsetof (struct _pci_config_regs, ROM_Base_Address), 1, (char *)&address);
		BaseReg->Address &= 0x7fe;
		}

	if (device->io_alloc.Address) {
		pci_free_resource (&device->io_alloc, device->BusNumber);
		}
	if (device->mem_alloc.Address) {
		pci_free_resource (&device->mem_alloc, device->BusNumber);
		}
	if (device->pref_alloc.Address) {
		pci_free_resource (&device->pref_alloc, device->BusNumber);
		}

	// Don't release the IRQ if this device is attached to a CardBus controller
	if (device->bmap->bdev->ClassCode == (PCI_CLASS_BRIDGE | PCI_SUBCLASS_BRIDGE_CARDBUS)) {
		return (1);		//Leave device enabled
		}

	if (device->IrqNumber != 0 && device->IrqNumber != 0xffffffff) {
		pci_free_irq (device->IrqNumber);
		}
	return (enabled);
}

/**************************************************************************/
/* release config of device                                               */
/**************************************************************************/

static	int		PCI_detach_device (void *handle, DEV_LIST *list, int close_flag)

{
Device		*device = DeviceHead;
int			i, status, enabled = 1;
uint32_t	flags = 0;

	status = PCI_DEVICE_NOT_FOUND;
	pthread_mutex_lock (&serv_mutex);
	while (device) {
		if (device == (Device *) handle) {
			if (device->LockFlg) {
				if (list && list->lockflg) {
					if (close_flag)
						i = (list->lockflg == device->LockFlg) ? list->lockflg : 1;
					else
						i = 1;
					device->LockFlg -= i;
					list->lockflg -= i;
					}
				else {
					device->LockFlg--;
					}
				}
			else {
				if (device->MgrAttach) {
					device->MgrAttach = 0;
					}
				}
			flags = device->AttachFlags & (PCI_INIT_ALL | PCI_INIT_ROM);
			if (! device->Persistent && flags && ! device->LockFlg && ! device->MgrAttach) {
				enabled = release_resources (device);
				}
			if (list && ! device->LockFlg) {
				for (i = 0; i < NUM_DEVICES; i++) {
					if (list->dev_list [i] == device) {
						list->dev_list [i] = NULL;
						list->num_devices--;
						}
					}
				}
			status = PCI_SUCCESS;
			break;
			}
		device = device->next;
		}
	if (status == PCI_SUCCESS && ! device->LockFlg && ! device->MgrAttach) {
		device->AttachFlags = 0;	//Clear attach flags if last detach
		}

	if (status == PCI_SUCCESS) {
		uint32_t	class = device->ClassCode & 0xff0000;

		if (class != PCI_CLASS_DISPLAY && class != PCI_CLASS_BRIDGE &&
			class != PCI_CLASS_PROCESSOR) {
			if (status == PCI_SUCCESS && !device->Persistent && !enabled) {
				off_pci (device->BusNumber, device->DevFuncNumber);
				}
			if ((device->PmState & 0x03) && !device->Persistent && flags)
				power_down_device (device->BusNumber, device->DevFuncNumber, device);
			}
		}
	pthread_mutex_unlock (&serv_mutex);
	return (status);
}

/**************************************************************************/
/* Setup I/O apertures in PCI-PCI bridges.                                */
/**************************************************************************/

static	int		setup_pci_io (uint64_t io_size)

{
rsrc_request_t	rreq;
int				allocd = 0, i, bit_32 = 0;
uint64_t		map_addr, size;
uint32_t		bus, devfn;
uint16_t		base, limit;
uint8_t			io_base [2];
bus_require_t	*req;
BUSMAP			*map;
Device			*dev;

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.start = 0;
	rreq.end = 0;
	rreq.length = io_size;
	rreq.align = KB4;
	rreq.flags = RSRCDBMGR_IO_PORT | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_TOPDOWN;
	for (i = 0; i < lastbus; i++) {
		req = &brqt [i];
		if (req->io_reqd == 0)
			continue;
		if ((map = find_bus (i)) == NULL)
			return (-1);
		if (map->io_start && map->io_end)	/* Bridge already setup */
			continue;
		if (! allocd) {
			if (rsrcdbmgr_attach (&rreq, 1) != EOK) {
				return (-1);
				}
			(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_IO_SPACE | PCI_MAP_ADDR_PCI);
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "CPU I/O %llx - PCI I/O %llx", rreq.start, map_addr);
			allocd = 1;
			}
		size = req->io_reqd;
		while (map) {
			dev = map->bdev;
			if (dev->ClassCode == host_bridge_class)
				break;
			bus = dev->BusNumber;
			devfn = dev->DevFuncNumber;
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Bridge Bus %x - devfunc %x", dev->BusNumber,
				dev->DevFuncNumber);
			if (map->io_start == 0) {
				map->io_start = map_addr;
				map->io_end = map_addr + (size - KB4 + (KB4 - 1));
				}
			else {
				if (map_addr > map->io_start) {
					map->io_end = map_addr + (size - KB4 + (KB4 - 1));
					}
				else {
					map->io_start = map_addr;
					}
				}

			if (PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Base),
				1, io_base)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "io_base0 %x - io_base1 %x", io_base [0], io_base [1]);
			if (io_base [0] & 1)	/* 32-bit addressing */
				bit_32 = 1;
			io_base [0] = ((map->io_start & 0xf000) >> 8);
			io_base [1] = ((map->io_end & 0xf000) >> 8);
			PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Base),
				1, io_base);
			if (bit_32) {
				PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Base_Upper16),
					1, &base);
				PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Limit_Upper16),
					1, &limit);
				my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "base %x - limit %x", base, limit);
				base = (map->io_start >> 16);
				limit = (map->io_end >> 16);
				PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Base_Upper16),
					1, &base);
				PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, IO_Limit_Upper16),
					1, &limit);
				}
			if (PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			base |= (PCI_COMMAND_IO_ENABLE | PCI_COMMAND_MASTER_ENABLE);
			PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base);
			map = map->prev;
			}
		map_addr += size;
		}
	if (allocd) {
		if (rsrcdbmgr_detach (&rreq, 1))
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "I/O detach failed");
		}

	return (0);
}

/**************************************************************************/
/* Setup prefetchable memory apertures in PCI-PCI bridges.                */
/**************************************************************************/

static	int		setup_pci_pref (uint64_t pref_size)

{
rsrc_request_t	rreq;
int				allocd = 0, i, bit_64 = 0;
uint64_t		map_addr, size;
uint32_t		bus, devfn, base, limit;
uint16_t		word [2];
bus_require_t	*req;
BUSMAP			*map;
Device			*dev;

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.start = 0;
	rreq.end = 0;
	rreq.length = pref_size;
	rreq.align = MEG1;
	rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_TOPDOWN;
	for (i = 0; i < lastbus; i++) {
		req = &brqt [i];
		if (req->pref_reqd == 0)
			continue;
		if ((map = find_bus (i)) == NULL)
			return (-1);
		if (map->prefetch_mem_start && map->prefetch_mem_end)	/* Bridge already setup */
			continue;
		if (! allocd) {
			if (rsrcdbmgr_attach (&rreq, 1) != EOK) {
				return (-1);
				}
			(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_MEM_SPACE | PCI_MAP_ADDR_PCI);
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "CPU memory %llx - PCI memory %llx", rreq.start, map_addr);
			allocd = 1;
			}
		size = req->pref_reqd;
		while (map) {
			dev = map->bdev;
			if (dev->ClassCode == host_bridge_class)
				break;
			bus = dev->BusNumber;
			devfn = dev->DevFuncNumber;
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Bridge Bus %x - devfunc %x", dev->BusNumber,
				dev->DevFuncNumber);
			if (map->prefetch_mem_start == 0) {
				map->prefetch_mem_start = map_addr;
				map->prefetch_mem_end = map_addr + (size - MEG1 + (MEG1 - 1));
				}
			else {
				if (map_addr > map->prefetch_mem_start) {
					map->prefetch_mem_end = map_addr + (size - MEG1 + (MEG1 - 1));
					}
				else {
					map->prefetch_mem_start = map_addr;
					}
				}

			if (PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Memory_Base),
				1, word)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "word0 %x - word1 %x", word [0], word [1]);
			if (word [0] & 1)	/* 64-bit addressing */
				bit_64 = 1;
			word [0] = ((map->prefetch_mem_start & 0xfff00000) >> 16);
			word [1] = ((map->prefetch_mem_end & 0xfff00000) >> 16);
			PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Memory_Base),
				1, word);
			if (bit_64) {
				PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Base_Upper32),
					1, &base);
				PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Limit_Upper32),
					1, &limit);
				my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "base %x - limit %x", base, limit);
				base = (map->prefetch_mem_start >> 32);
				limit = (map->prefetch_mem_end >> 32);
				PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Base_Upper32),
					1, &base);
				PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Prefetchable_Limit_Upper32),
					1, &limit);
				}
			if (PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			base |= (PCI_COMMAND_MEM_ENABLE | PCI_COMMAND_MASTER_ENABLE);
			PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base);
			map = map->prev;
			}
		map_addr += size;
		}
	if (allocd) {
		if (rsrcdbmgr_detach (&rreq, 1))
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Pref detach failed");
		}

	return (0);
}


/**************************************************************************/
/* Setup memory apertures in PCI-PCI bridges.                             */
/**************************************************************************/

static	int		setup_pci_mem (uint64_t mem_size)

{
rsrc_request_t	rreq;
int				allocd = 0, i;
uint64_t		map_addr, size;
uint32_t		bus, devfn;
uint16_t		base, word [2];
bus_require_t	*req;
BUSMAP			*map;
Device			*dev;

	memset ((char *) &rreq, 0, sizeof (rreq));
	rreq.start = 0;
	rreq.end = 0;
	rreq.length = mem_size;
	rreq.align = MEG1;
	rreq.flags = RSRCDBMGR_PCI_MEMORY | RSRCDBMGR_FLAG_ALIGN | RSRCDBMGR_FLAG_TOPDOWN;
	for (i = 0; i < lastbus; i++) {
		req = &brqt [i];
		if (req->mem_reqd == 0)
			continue;
		if ((map = find_bus (i)) == NULL)
			return (-1);
		if (map->mem_start && map->mem_end)	/* Bridge already setup */
			continue;
		if (! allocd) {
			if (rsrcdbmgr_attach (&rreq, 1) != EOK) {
				return (-1);
				}
			(*pci_drivers->entry->map_addr) (ll_handle, rreq.start, &map_addr, PCI_MEM_SPACE | PCI_MAP_ADDR_PCI);
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "CPU memory %llx - PCI memory %llx", rreq.start, map_addr);
			allocd = 1;
			}
		size = req->mem_reqd;
		while (map) {
			dev = map->bdev;
			if (dev->ClassCode == host_bridge_class)
				break;
			bus = dev->BusNumber;
			devfn = dev->DevFuncNumber;
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "Bridge Bus %x - devfunc %x", dev->BusNumber,
				dev->DevFuncNumber);
			if (map->mem_start == 0) {
				map->mem_start = map_addr;
				map->mem_end = map_addr + (size - MEG1 + (MEG1 - 1));
				}
			else {
				if (map_addr > map->mem_start) {
					map->mem_end = map_addr + (size - MEG1 + (MEG1 - 1));
					}
				else {
					map->mem_start = map_addr;
					}
				}

			if (PCI_read_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Memory_Base),
				1, word)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "word0 %x - word1 %x", word [0], word [1]);
			word [0] = ((map->mem_start & 0xfff00000) >> 16);
			word [1] = ((map->mem_end & 0xfff00000) >> 16);
			PCI_write_config32 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Memory_Base),
				1, word);
			if (PCI_read_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base)) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Read config error");
				return (-1);
				}
			base |= (PCI_COMMAND_MEM_ENABLE | PCI_COMMAND_MASTER_ENABLE);
			PCI_write_config16 (bus, devfn, offsetof (struct _pci_bridge_config_regs, Command),
				1, &base);
			map = map->prev;
			}
		map_addr += size;
		}
	if (allocd) {
		if (rsrcdbmgr_detach (&rreq, 1))
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Mem detach failed");
		}

	return (0);
}

/**************************************************************************/
/* Initialise all variables and scan the PCI bus(ses) for devices.        */
/**************************************************************************/

static	int		PCI_Startup (void)

{
int			i, status;
Device		*device;
pci_bus_t	*bus;
BUSMAP		*map, *peer;
bus_require_t	*req;
uint64_t	io_size, mem_size, pref_size;
#ifdef	__X86__
uint32_t	tmp;

	(*pci_drivers->entry->version) (ll_handle, &tmp, &version, &hardware);
#else
	version = 0x210;
	hardware = 0;
#endif

	lastbus = 0;
	numdevices = 0;
	listcount = 0;
	DeviceHead = DeviceTail = NULL;
	io_size = mem_size = pref_size = 0;

	find_ranges ();

	if (scan_all () != PCI_SUCCESS || lastbus == 0) {		//No PCI found!
		return (-1);
		}

	/* Scan through all the devices and configure the bridges */

	for (i = 0; i < lastbus; i++) {
		req = &brqt [i];
		if (req->io_reqd) {
			if (req->io_reqd < KB4) {
				io_size += KB4;
				req->io_reqd = KB4;
				}
			else {
				if (req->io_reqd % KB4) {
					req->io_reqd += KB4;
					req->io_reqd &= ~(KB4 - 1);
					}
				io_size += req->io_reqd;
				}
			}
		if (req->mem_reqd) {
			if (req->mem_reqd < MEG1) {
				mem_size += MEG1;
				req->mem_reqd = MEG1;
				}
			else {
				if (req->mem_reqd % MEG1) {
					req->mem_reqd += MEG1;
					req->mem_reqd &= ~(MEG1 - 1);
					}
				mem_size += req->mem_reqd;
				}
			}
		if (req->pref_reqd) {
			if (req->pref_reqd < MEG1) {
				pref_size += MEG1;
				req->pref_reqd = MEG1;
				}
			else {
				if (req->pref_reqd % MEG1) {
					req->pref_reqd += MEG1;
					req->pref_reqd &= ~(MEG1 - 1);
					}
				pref_size += req->pref_reqd;
				}
			}
		}
	if (verbose > 2)
		my_slogf (_SLOGC_PCI, _SLOG_DEBUG1, "I/O reqd %llx - Mem reqd %llx - Pref reqd %llx",
			io_size, mem_size, pref_size);

	if (io_size) {
		if (setup_pci_io (io_size))
			return (-1);
		}
	if (mem_size) {
		if (setup_pci_mem (mem_size))
			return (-1);
		}

	if (pref_size) {
		if (setup_pci_pref (pref_size))
			return (-1);
		}

	device = DeviceHead;
	while (device) {
		if (((device->ClassCode & 0xf0000) == PCI_CLASS_BRIDGE) && (device->DevFuncNumber & 7) == 0) {
			if ((map = find_bus (device->bmap->secondary_bus)) == NULL)
				return (-1);
			bus = map->pci_bus;
			req = &brqt [map->secondary_bus];
			if (verbose > 2) {
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Busmap Pri %x - Sec %x - Sub %x", map->primary_bus,
					map->secondary_bus, map->subordinate_bus);
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "I/O Start %llx - I/O End %llx", map->io_start,
					map->io_end);
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Mem Start %llx - Mem End %llx", map->mem_start,
					map->mem_end);
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "I/O %llx - Mem %llx - Pref %llx", req->io_reqd,
					req->mem_reqd, req->pref_reqd);
				}
			if ((status = (*pci_drivers->entry->cnfg_bridge) (ll_handle,
				device->BusNumber, device->DevFuncNumber, bus)) != PCI_SUCCESS) {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Unable to configure bridge");
				return (status);
				}
			}
		device = device->next;
		}

	map = bus_map;
	while (map) {
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Map %p - Next %p - Prev %p - Peer %p", 
			map, map->next, map->prev, map->peer);
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Prim %x - Sec %x - Sub %x",
			map->primary_bus, map->secondary_bus, map->subordinate_bus);
		if (map->peer) {
			peer = map->peer;
			while (peer) {
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Peer %p - Next %p - Prev %p - Peer %p", 
					peer, peer->next, peer->prev, peer->peer);
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "Prim %x - Sec %x - Sub %x",
					peer->primary_bus, peer->secondary_bus, peer->subordinate_bus);
				peer = peer->peer;
				}
			}
		map = map->next;
		}

	return (PCI_SUCCESS);
}

/**************************************************************************/
/* Allocate an ocb for this open, attach a device list buffer to it.      */
/**************************************************************************/

IOFUNC_OCB_T	*PCI_OcbCalloc (resmgr_context_t *ctp, iofunc_attr_t *attr)

{
DEV_LIST		*ocb;

	if (verbose)
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "io_open for id = %x", ctp->rcvid);

	if ((ocb = calloc (1, sizeof (*ocb))) == NULL) {
		if (verbose) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "couldn't allocate %d bytes", sizeof (*ocb));
			}
		return (NULL);
		}

	return (ocb);
}

/**************************************************************************/
/* Free the ocb allocated above as well as the device list buffer.        */
/**************************************************************************/

void	PCI_OcbFree (IOFUNC_OCB_T *list)

{
int		i;

	if (verbose) {
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "io_close_ocb");
		}

	for (i = 0; i < NUM_DEVICES; i++) {
		if (list->dev_list [i]) {
			PCI_detach_device (list->dev_list [i], list, 1);
			}
		}

	free (list);
}

/**************************************************************************/
/* Handle user I/O messages.                                              */
/**************************************************************************/

int		PCI_IoMsg (resmgr_context_t *ctp, io_msg_t *msg, IOFUNC_OCB_T *ocb, iofunc_attr_t *attr)

{
union _pci_data *data = ((void *) (sizeof (msg->i) + sizeof (struct _pci_reply_hdr) + (char *)(&msg->i)));
uint32_t		t1, t2, t3, devfn;
uint8_t			bus;
int				ret, reply_length = 0;
int				local_errno, type;
Device			*handle, *device = DeviceHead;
int				(*rout)(uint8_t, uint32_t, uint32_t, uint32_t, void *);
struct			_pci_message	msg_reply;

	if (msg->i.mgrid != _IOMGR_PCI)
		return (ENOSYS);

	local_errno = EOK;
	memcpy ((char *) &msg_reply, (char *) &msg->i, msg->i.combine_len);
//	if (verbose > 2)
//		my_slogf (_SLOGC_PCI, _SLOG_INFO, "Message type %d - id = %x", msg->i.subtype, ctp->rcvid);
	switch (msg->i.subtype) {
		case	IOM_PCI_READ_CONFIG_BYTE:
			ret = PCI_read_config8 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char *)&msg_reply.msg.config.buffer);
			reply_length += sizeof (struct _pci_config);
			break;
		case	IOM_PCI_READ_CONFIG_WORD:
			ret = PCI_read_config16 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char *)&msg_reply.msg.config.buffer);
			reply_length += sizeof (struct _pci_config);
			break;
		case	IOM_PCI_READ_CONFIG_DWORD:
			ret = PCI_read_config32 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char *)&msg_reply.msg.config.buffer);
			reply_length += sizeof (struct _pci_config);
			break;
		case	IOM_PCI_WRITE_CONFIG_BYTE:
			ret = PCI_write_config8 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char*)&data->config.buffer);
			break;
		case	IOM_PCI_WRITE_CONFIG_WORD:
			ret = PCI_write_config16 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char*)&data->config.buffer);
			break;
		case	IOM_PCI_WRITE_CONFIG_DWORD:
			ret = PCI_write_config32 (data->config.bus, data->config.devfunc, data->config.offset, data->config.count, (char*)&data->config.buffer);
			break;
		case	IOM_PCI_READ_CONFIG_HANDLE:
		case	IOM_PCI_WRITE_CONFIG_HANDLE:
			type = msg->i.subtype == IOM_PCI_READ_CONFIG_HANDLE ? 0 : 1;
			handle = (Device *) data->config_hdl.handle;
			rout = NULL;
			ret = PCI_DEVICE_NOT_FOUND;
			while (device) {
				if (device == handle) {
					bus = device->BusNumber;
					devfn = device->DevFuncNumber;
					switch (data->config_hdl.size) {
						case	1:		//Byte
							rout = type ? PCI_write_config8 : PCI_read_config8;
							break;
						case	2:		//Word
							rout = type ? PCI_write_config16 : PCI_read_config16;
							break;
						case	4:		//DWord
							rout = type ? PCI_write_config32 : PCI_read_config32;
							break;
						default:
							rout = NULL;
							break;
						}
					if (rout) {
						ret = (*rout)(bus, devfn, data->config_hdl.offset,
							data->config_hdl.count, (char *)&msg_reply.msg.config_hdl.buffer);
						reply_length += sizeof (struct _pci_config_hdl);
						}
					break;
					}
				device = device->next;
				}
			break;
		case	IOM_PCI_FIND_DEVICE:
			ret = PCI_find_device (data->device.deviceid, data->device.vendorid, data->device.index, &bus, &t2);
			msg_reply.msg.device.busnum = bus;
			msg_reply.msg.device.devfuncnum = t2;
			reply_length += sizeof (struct _pci_device);
			break;
		case	IOM_PCI_FIND_CLASS:
			ret = PCI_find_class (data->class.class, data->class.index, &bus, &t2);
			msg_reply.msg.class.busnum = bus;
			msg_reply.msg.class.devfuncnum = t2;
			reply_length += sizeof (struct _pci_class);
			break;
		case	IOM_PCI_BIOS_PRESENT:
			ret = pci_bios_present (&t1, &t2, &t3);
			msg_reply.msg.present.lastbus = t1;
			msg_reply.msg.present.version = t2;
			msg_reply.msg.present.hardware = t3;
			reply_length += sizeof (struct _pci_present);
			break;
		case	IOM_PCI_MAP_ADDRESS: {
			uint64_t	addr;

			ret = (*pci_drivers->entry->map_addr) (ll_handle, data->map.pciaddr, &addr, data->map.flags);
			msg_reply.msg.map.cpuaddr = addr;
			reply_length += sizeof (struct _pci_map);
			}
			break;
		case	IOM_PCI_GET_VERSION:
			ret = PCI_GetVersion (&msg_reply.msg.version.version, &msg_reply.msg.version.numdevices);
			reply_length += sizeof (struct _pci_version);
			break;
		case	IOM_PCI_ATTACH_DEVICE:
			msg_reply.msg.lock.handle = PCI_attach_device (data->lock.handle,
				data->lock.flags, data->lock.idx, &msg_reply.msg.lock.configbuffer,
				ocb, &local_errno);
			reply_length += msg_reply.msg.lock.handle ? sizeof (struct _pci_attach) : 0;
			ret = reply_length ? PCI_SUCCESS : PCI_DEVICE_NOT_FOUND;
			break;
		case	IOM_PCI_DETACH_DEVICE:
			ret = PCI_detach_device (data->unlock.handle, ocb, 0);
			break;
		case	IOM_PCI_MAP_IRQ:
			ret = (*pci_drivers->entry->map_irq) (ll_handle, data->map_irq.bus,
					data->map_irq.devfunc, data->map_irq.intno, data->map_irq.intpin,
					PCI_MAP_IRQ);
			break;
		case	IOM_PCI_RESCAN_BUS:
			ret = rescan ();
			break;
		case	IOM_PCI_IRQ_ROUTING_OPTIONS:
			ret = (*pci_drivers->entry->routing_options) ((struct _pci_route_buffer *) &msg_reply.msg.route_opt.buf_size,
					(uint16_t *) &msg_reply.msg.route_opt.irq_info);
			reply_length += (ret == PCI_SUCCESS) ? sizeof (struct _pci_route_opt) : 0;
			break;
		default:
			local_errno = ENOSYS;
			ret = PCI_SUCCESS;
			break;
		}
	if (local_errno != EOK)
		return (local_errno);
	_RESMGR_STATUS (ctp, local_errno);
	msg_reply.rep_hdr.reply_status = ret;
	reply_length += sizeof (struct _pci_reply_hdr);
	msg_reply.rep_hdr.reply_length = reply_length;
	msg_reply.msg_hdr.i.combine_len = reply_length + sizeof (io_msg_t);
	SETIOV (ctp->iov, &msg_reply, reply_length + sizeof (io_msg_t));
	MsgReplyv (ctp->rcvid, ret, ctp->iov, 1);
	return (_RESMGR_NOREPLY);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

int		pci_dll_load (char *options, char *optarg)

{
void			*d;
int8_t			dllpath [_POSIX_PATH_MAX + 1];
pdrvr_entry_t	*entry;
pdrvr_t			*pdrvr;

const struct dll_list	*l;

	if (options == NULL && optarg == NULL) {
		for (l = dll_list; l->fname != NULL; ++l) {
			if (l->fname) {
				sprintf (dllpath, "%s", l->fname);
				break;
				}
			}
		}
	else {
		if (strchr (optarg, '/') != NULL) {
			strcpy (dllpath, optarg);
			}
		else {
			sprintf (dllpath, "pci-%s.so", optarg);
			}
		}

	if (pdrvr = calloc (1, sizeof (pdrvr_t))) {
		if (d = pci_dlopen (dllpath, 0)) {
			if (entry = (void *) pci_dlsym (d, "pci_entry")) {
				if (!(entry->attach)(options, &pdrvr->hdl)) {
					ll_handle = pdrvr->hdl;
					if (pci_drivers) {
						pdrvr->next = pci_drivers;
						pci_drivers->next = pdrvr;
						}
					pci_drivers = pdrvr;
					pdrvr->entry = entry;
					return (EOK);
					}
				}
			else {
				my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s dlsym failed %s", progname, optarg);
				}
			pci_dlclose (d);
			}
		else {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s unable to load %s", progname, dllpath);
			}
		}
	else {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s Unable to allocate PCI device memory", progname);
		}
		
	return (ENODEV);
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

void	pci_signal_handler (int signo)

{
pdrvr_t		*pdrvr;

	if (signo == SIGTERM) {
		// for each driver call detach
		pdrvr = pci_drivers;
		for (; pdrvr; pdrvr = pdrvr->next) {
			if (pdrvr->entry->detach) {
				my_slogf (_SLOGC_PCI, _SLOG_INFO, "%s shutting down pci %lx", progname, (ulong_t) pdrvr->hdl);
				pdrvr->entry->detach (pdrvr->hdl);
				}
			}
		exit (0);
		}
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

void	pci_resmgr_loop (resmgr_context_t *ctp)

{
	while(1) {
		if ((resmgr_block (ctp)) == NULL) {
			my_slogf (_SLOGC_PCI, _SLOG_ERROR, "block error");
			exit (EXIT_FAILURE);
			}
		resmgr_handler (ctp);
		}
}

/**************************************************************************/
/*                                                                        */
/**************************************************************************/

int		main (int argc, char **argv)

{
int				c, dll_loaded = 0;
iofunc_attr_t	io_attr;
resmgr_connect_funcs_t connect_funcs;
dispatch_t 		*dpp;
resmgr_attr_t 	resmgr_attr;
resmgr_context_t *ctp;
iofunc_mount_t	io_mount;
iofunc_funcs_t	io_ocb_funcs;

	pci_drivers = NULL;
	scan_buses = MAX_BUSES;

	// Enable IO capability.
	if (ThreadCtl (_NTO_TCTL_IO, NULL) == -1) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s ThreadCtl", progname);
		return (EPERM);
		}

	if((dpp = dispatch_create()) == NULL) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s: Unable to allocate dispatch context.",argv[0]);
		exit (EXIT_FAILURE);
		}

	memset(&resmgr_attr, 0, sizeof resmgr_attr);
	resmgr_attr.nparts_max = 20;
	resmgr_attr.msg_max_size = 0;

	// process args	

	while (optind < argc) {
		if ((c = getopt (argc, argv, "b:Bd:hmpvx")) == -1) {
			optind++;
			continue;
			}

		switch (c) {
			case	'b':
				scan_buses = atoi (optarg);
				break;
			case	'B':
				enum_xtrans_bridge = 1;
				break;

			case	'd':
				pci_dll_load (argv [optind], optarg);
				dll_loaded = 1;
				break;

			case	'h':
				hw_override = 1;
				break;

			case	'm':
				map_irqs ^= 1;		//Flip the flag
				break;
			case	'p':
				program_display_device = 1;
				break;
			case	'v':
				verbose++;
				break;

			case	'x':
				dont_touch = 1;
				break;

			default:
				my_slogf (_SLOGC_PCI, _SLOG_WARNING, "%s unknown option", progname);
				break;
			}
		}

	if (verbose) {
		my_slogf (_SLOGC_PCI, _SLOG_INFO, "%s %s %s", progname, __DATE__, __TIME__);
		}

	if (! dll_loaded)
		pci_dll_load (NULL, NULL);

	if (!pci_drivers) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s No dll loaded!", progname);
		exit (EXIT_FAILURE);
		}

	memset ((char *) &io_mount, 0, sizeof (io_mount));
	memset ((char *) &io_ocb_funcs, 0, sizeof (iofunc_funcs_t));
	io_ocb_funcs.nfuncs = _IOFUNC_NFUNCS;
	io_ocb_funcs.ocb_calloc = &PCI_OcbCalloc;
	io_ocb_funcs.ocb_free = &PCI_OcbFree;
	io_mount.funcs = &io_ocb_funcs;
	iofunc_attr_init (&io_attr, 0600 | S_IFCHR, NULL, NULL);
	io_attr.mount = &io_mount;
	iofunc_func_init (_RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS, &io_funcs);
	io_funcs.msg = (void *) &PCI_IoMsg;

	if(resmgr_attach (dpp, &resmgr_attr, "/dev/pci", _FTYPE_ANY, 0, &connect_funcs, &io_funcs, &io_attr) == -1) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "resmgr_pathname_attach failed");
		exit (EXIT_FAILURE);
		}

	if((ctp = resmgr_context_alloc(dpp)) == NULL) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "Could not allocate resmgr context");
		exit (EXIT_FAILURE);
		}
	
	// initialize configuration manager
	if (PCI_Startup () != PCI_SUCCESS) {
		my_slogf (_SLOGC_PCI, _SLOG_ERROR, "%s PCI_Startup failed!", progname);
		exit (EXIT_FAILURE);
		}

#if 0 	/* For debugging, keep around stdout/stderr */
	procmgr_daemon(EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE);
#else
	procmgr_daemon(EXIT_SUCCESS, PROCMGR_DAEMON_NODEVNULL | PROCMGR_DAEMON_NOCLOSE);
#endif

	signal (SIGTERM, pci_signal_handler);
	pci_resmgr_loop (ctp);

	exit (EXIT_SUCCESS);
}

